from __future__ import annotations

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import triton
import triton.language as tl
import textwrap
import inspect
import random
from typing import Sequence, Dict, Optional
from triton.runtime.jit import JITFunction
from copy import deepcopy

# In the latest triton, math functions were shuffled around into different modules:
# https://github.com/openai/triton/pull/3172
if hasattr(tl.extra.cuda, "libdevice"):
    tlmath = tl.extra.cuda.libdevice
else:
    tlmath = tl.math

from pyjuice.nodes import InputNodes
from pyjuice.utils.grad_fns import ReverseGrad
from pyjuice.utils import BitSet
from pyjuice.utils.source2fn import make_function_from_src
from pyjuice.utils.kernel_launcher import triton_jit
from .layer import Layer


class InputLayer(Layer, nn.Module):
    def __init__(self, nodes: Sequence[InputNodes], cum_nodes: int = 0, pc_num_vars: int = 0, max_tied_ns_per_parflow_block: int = 4) -> None:
        """
        Compiler flags:
        - `max_tied_ns_per_parflow_block`: the maximum number of tied nodes allowed in the backward pass. Setting to a larger value will
                                           lead to reduced memory overhead but might lead to additional computational burden due to conflicts
                                           in gradient accumulation.
        """

        assert len(nodes) == len(set(nodes)), "Input node list contains duplicates."

        # Reorder input nodes such that for any tied nodes, its source nodes appear before them
        nodes = self._reorder_nodes(nodes)

        nn.Module.__init__(self)
        Layer.__init__(self, nodes, disable_block_size_check = True)

        # Total number of variables
        self.pc_num_vars = pc_num_vars

        ## Parse input `nodes` ##
        node_vars = []
        node_sizes = []
        node_metadata = []
        layer_num_nodes = 0
        cum_params = 0
        cum_param_flows = 0
        cum_source_ns = 0
        dist_signature = None
        node2tiednodes = dict()
        for node_id, ns in enumerate(self.nodes):
            if dist_signature is None:
                dist_signature = ns.dist.get_signature()
            else:
                assert dist_signature == ns.dist.get_signature(), "Nodes of an InputLayer must have the same distribution type."

            node_vars.append(ns.scope.to_list())
            node_sizes.append(ns.num_nodes)
            node_metadata.append(ns.dist.get_metadata())

            ns._output_ind_range = (cum_nodes, cum_nodes + ns.num_nodes)
            cum_nodes += ns.num_nodes
            layer_num_nodes += ns.num_nodes

            if not ns.is_tied():
                cum_params += ns.num_nodes * ns.dist.num_parameters()
                ns._param_range = (cum_params - ns.num_nodes * ns.dist.num_parameters(), cum_params)

                cum_param_flows += ns.num_nodes * ns.dist.num_param_flows()
                ns._param_flow_range = (cum_param_flows - ns.num_nodes * ns.dist.num_param_flows(), cum_param_flows)

                cum_source_ns += ns.num_nodes
            else:
                source_ns = ns.get_source_ns()
                ns._param_range = deepcopy(source_ns._param_range)

                if source_ns not in node2tiednodes:
                    node2tiednodes[source_ns] = [[source_ns], 1, source_ns._param_flow_range]
                
                dup_count = node2tiednodes[source_ns][1]
                if dup_count >= max_tied_ns_per_parflow_block:
                    cum_param_flows += ns.num_nodes * ns.dist.num_param_flows()
                    ns._param_flow_range = (cum_param_flows - ns.num_nodes * ns.dist.num_param_flows(), cum_param_flows)
                    node2tiednodes[source_ns][2] = ns._param_flow_range

                    node2tiednodes[source_ns][0].append(ns)
                    node2tiednodes[source_ns][1] = 1
                else:
                    ns._param_flow_range = deepcopy(node2tiednodes[source_ns][2])

                    node2tiednodes[source_ns][1] += 1

        self._output_ind_range = (cum_nodes - layer_num_nodes, cum_nodes)
        self.num_parameters = cum_params
        self.num_param_flows = cum_param_flows
        self.num_nodes = layer_num_nodes
        self.dist_signature = dist_signature

        # Store the triton kernel functions implemented by the target `Distribution`
        self.fw_mar_fn = self.nodes[0].dist.get_fw_mar_fn()
        self.bk_flow_fn = self.nodes[0].dist.get_bk_flow_fn()
        self.sample_fn = self.nodes[0].dist.get_sample_fn()
        self.em_fn = self.nodes[0].dist.get_em_fn()

        try:
            self.post_fw_fns = self.nodes[0].dist.post_fw_fns
        except AttributeError:
            self.post_fw_fns = []

        try:
            self.post_bp_fns = self.nodes[0].dist.post_bp_fns
        except AttributeError:
            self.post_bp_fns = []

        if hasattr(self.nodes[0].dist, "bk_flow_mask_fn"):
            self.bk_flow_mask_fn = self.nodes[0].dist.bk_flow_mask_fn
        else:
            self.bk_flow_mask_fn = None
        
        if hasattr(self.nodes[0].dist, "partition_fn"):
            self.fw_partition_fn = self.nodes[0].dist.partition_fn
        else:
            self.fw_partition_fn = None

        ## Prepair and compile the layer ##
        num_vars = len(node_vars[0])
        # Start variable index: vids[i,:] are the variables of the ith node
        vids = torch.empty([self.num_nodes, num_vars], dtype = torch.long)
        # Start parameter index: params[s_pids[i]] is the first parameter of the ith node
        s_pids = torch.empty([self.num_nodes], dtype = torch.long)
        # Start parameter flow index: param_flows[s_pfids[i]] is the first parameter flow of the ith node
        s_pfids = torch.empty([self.num_nodes], dtype = torch.long)
        # Start metadata index: metadata[s_mids[i]] is the first metadata of the ith node
        metadata = []
        s_mids = torch.empty([self.num_nodes], dtype = torch.long)
        # Node local index: s_nids[i] is the local id of this node within all nodes defined on the same variable
        nids = torch.empty([self.num_nodes], dtype = torch.long)
        # Source node ids (nodes with their original parameters)
        source_nids = torch.empty([cum_source_ns], dtype = torch.long)
        # Pointer to a node's source id
        node_id2source_id = torch.empty([self.num_nodes], dtype = torch.long)

        # Parameters of this layer
        params = torch.empty([max(self.num_parameters, 1)], dtype = torch.float32)
        
        n_start = 0
        source_n_start = 0
        var2count = dict()
        ns2n_range = dict()
        for ns_id, ns in enumerate(self.nodes):
            n_end = n_start + ns.num_nodes
            ns2n_range[ns] = (n_start, n_end)

            # `vids`
            assert len(node_vars[ns_id]) == num_vars
            vids[n_start:n_end,:] = torch.tensor(node_vars[ns_id]).view(1, -1)

            # `s_pids` and `s_pfids`
            if ns.dist.num_parameters() > 0:
                pid_offsets = torch.arange(0, ns.num_nodes * ns.dist.num_parameters(), ns.dist.num_parameters())
                s_pids[n_start:n_end] = ns._param_range[0] + pid_offsets
            else:
                s_pids[n_start:n_end] = 0

            if ns.dist.num_param_flows() > 0:
                pfid_offsets = torch.arange(0, ns.num_nodes * ns.dist.num_param_flows(), ns.dist.num_param_flows())
                s_pfids[n_start:n_end] = ns._param_flow_range[0] + pfid_offsets
            else:
                s_pfids[n_start:n_end] = 0

            # `source_nids`
            if not ns.is_tied():
                source_n_end = source_n_start + ns.num_nodes
                source_nids[source_n_start:source_n_end] = torch.arange(n_start, n_end)
                source_n_start = source_n_end

            # `node_id2source_id`
            if ns.is_tied():
                temp_n_start, temp_n_end = ns2n_range[ns._source_node]
                node_id2source_id[n_start:n_end] = torch.arange(temp_n_start, temp_n_end)
            else:
                node_id2source_id[n_start:n_end] = -1 # Set to -1 as a flag

            # `metadata` and `s_mids`
            s_mids[n_start:n_end] = len(metadata)
            metadata.extend(node_metadata[ns_id])

            # `s_nids`
            if len(node_vars[ns_id]) == 1:
                var = node_vars[ns_id][0]
                if var not in var2count:
                    var2count[var] = 0
                nids[n_start:n_end] = torch.arange(0, n_end - n_start) + var2count[var]
                var2count[var] += n_end - n_start

            n_start = n_end

        self.register_buffer("vids", vids)
        self.register_buffer("s_pids", s_pids)
        self.register_buffer("s_pfids", s_pfids)
        self.register_buffer("metadata", torch.tensor(metadata).float())
        self.register_buffer("s_mids", s_mids)
        self.register_buffer("nids", nids)
        self.register_buffer("source_nids", source_nids)
        self.register_buffer("node_id2source_id", node_id2source_id)

        ## Prepare info buffers for tied nodes ##
        self.tied2source_nids = []
        for source_ns, item in node2tiednodes.items():
            if len(item[0]) > 1: # If the length is 1, then everything is already accumulated in the source node's parflow
                num_par_flows = source_ns._param_flow_range[1] - source_ns._param_flow_range[0]
                pfid_start = source_ns._param_flow_range[0]
                ch_nodes = item[0]

                ch_pfids = torch.empty([len(ch_nodes)], dtype = torch.long)
                for ch_id, ch_ns in enumerate(ch_nodes):
                    ch_pfids[ch_id] = ch_ns._param_flow_range[0]

                self.tied2source_nids.append([pfid_start, num_par_flows, ch_pfids])

        self.params = nn.Parameter(params)
        # Due to the custom inplace backward pass implementation, we do not track 
        # gradient of PC parameters by PyTorch.
        self.params.requires_grad = False

        self.num_vars_per_node = num_vars

        self.param_flows = None

        self.device = torch.device("cpu")

        self._used_external_params = False

        # Batch size of parameters in the previous forward pass
        self._param_batch_size = 1
    
    def to(self, device):
        nn.Module.to(self, device = device)

        # Take special care to `tied2source_nids`
        for i in range(len(self.tied2source_nids)):
            self.tied2source_nids[i][2] = self.tied2source_nids[i][2].to(device)

        self.device = device

    def init_param_flows(self, flows_memory: float = 1.0):
        batch_size = self._param_batch_size
        if self.param_flows is None \
                or (self.param_flows.dim() == 1 and batch_size > 1) \
                or (self.param_flows.dim() == 2 and batch_size != self.param_flows.size(1)):
            if batch_size == 1:
                shape = [self.num_param_flows]
            else:
                shape = [self.num_param_flows, batch_size]
            self.param_flows = torch.zeros(shape, device = self.device)
        else:
            assert self.param_flows.size(0) == self.num_param_flows
            if flows_memory < 1.0:
                self.param_flows[:] *= flows_memory

        return None

    def forward(self, data: torch.Tensor, node_mars: torch.Tensor, params: Optional[Dict] = None,
                missing_mask: Optional[torch.Tensor] = None, _batch_first: bool = True, 
                _apply_missing_mask_only: bool = False, **kwargs):
        self._used_external_params = (params is not None)

        if params is None:
            params = self.params
        else:
            params = params["params"]

        assert params.dim() == 1

        if "cuda" in self.device.type:
            # Need to flatten data to ensure the memory is aligned following [num_vars, batch_size]
            data = data.reshape(-1).contiguous()

            tot_num_nodes = node_mars.size(0)
            batch_size = node_mars.size(1)
            node_offset = self._output_ind_range[0]

            if not self.provided("fw_local_ids"):
                layer_num_nodes = self._output_ind_range[1] - self._output_ind_range[0]
                fw_local_ids = None
            else:
                layer_num_nodes = self.fw_local_ids.size(0)
                fw_local_ids = self.fw_local_ids

            if not self.provided("_mars_kernel"):
                self._mars_kernel = self._compile_triton_kernel(self._mars_kernel_template, mar_fn = self.fw_mar_fn)

            BLOCK_SIZE = 1024

            grid = (triton.cdiv(layer_num_nodes * batch_size, BLOCK_SIZE),)

            if not _apply_missing_mask_only:
                self._mars_kernel[grid](
                    params_ptr = self.params, 
                    node_mars_ptr = node_mars, 
                    data_ptr = data, 
                    vids_ptr = self.vids, 
                    s_pids_ptr = self.s_pids, 
                    metadata_ptr = self.metadata, 
                    s_mids_ptr = self.s_mids, 
                    fw_local_ids_ptr = fw_local_ids,
                    layer_num_nodes = layer_num_nodes, 
                    batch_size = batch_size, 
                    num_vars_per_node = self.num_vars_per_node, 
                    nv_block_size = triton.next_power_of_2(self.num_vars_per_node),
                    node_offset = node_offset, 
                    BLOCK_SIZE = BLOCK_SIZE, 
                    partial_eval = 1 if fw_local_ids is not None else 0,
                    num_warps = 8
                )
            else:
                assert missing_mask is not None, "`missing_mask` should be provided when `_apply_missing_mask_only = True`."

            # Apply missing mask if required
            if missing_mask is not None:
                assert self.num_vars_per_node == 1, "`missing_mask` only supported for univariate distributions."

                mask_dim = missing_mask.dim()
                if mask_dim == 1:
                    mode = 0
                elif _batch_first or self.pc_num_vars == missing_mask.size(1):
                    mode = 1
                else:
                    mode = 2

                grid = (triton.cdiv(layer_num_nodes * batch_size, BLOCK_SIZE),)

                self._fw_missing_mask_kernel[grid](
                    missing_mask_ptr = missing_mask,
                    node_mars_ptr = node_mars, 
                    vids_ptr = self.vids, 
                    fw_local_ids_ptr = fw_local_ids,
                    num_vars = self.pc_num_vars,
                    layer_num_nodes = layer_num_nodes, 
                    batch_size = batch_size, 
                    node_offset = node_offset, 
                    BLOCK_SIZE = 1024, 
                    partial_eval = 1 if fw_local_ids is not None else 0,
                    mode = mode,
                    num_warps = 8
                )

            # Apply post-processing kernels
            for (kernel, cond_fn, prep_kwargs_fn) in self.post_fw_fns:
                if not cond_fn(self, kwargs):
                    continue

                target_kwargs = prep_kwargs_fn(self, kwargs)

                BLOCK_SIZE = 1024

                grid = (triton.cdiv(layer_num_nodes * batch_size, BLOCK_SIZE),)

                kernel[grid](
                    params_ptr = self.params, 
                    node_mars_ptr = node_mars, 
                    data_ptr = data, 
                    vids_ptr = self.vids, 
                    s_pids_ptr = self.s_pids, 
                    metadata_ptr = self.metadata, 
                    s_mids_ptr = self.s_mids, 
                    nids_ptr = self.nids,
                    fw_local_ids_ptr = fw_local_ids,
                    layer_num_nodes = layer_num_nodes, 
                    batch_size = batch_size, 
                    num_vars_per_node = self.num_vars_per_node, 
                    nv_block_size = triton.next_power_of_2(self.num_vars_per_node),
                    node_offset = node_offset, 
                    BLOCK_SIZE = BLOCK_SIZE, 
                    partial_eval = 1 if fw_local_ids is not None else 0,
                    num_warps = 8,
                    **target_kwargs
                )

        else:
            raise NotImplementedError("CPU forward fn for input nodes is not implemented.")

    def backward(self, data: torch.Tensor, node_flows: torch.Tensor, 
                 node_mars: torch.Tensor, params: Optional[Dict] = None,
                 logspace_flows: bool = False, missing_mask: Optional[torch.Tensor] = None, 
                 _batch_first: bool = True, **kwargs):
        """
        data: [num_vars, B]
        node_flows: [num_nodes, B]
        node_mars: [num_nodes, B]
        """

        if params is None:
            params = self.params
        else:
            params = params["params"]

        assert params.dim() == 1

        if "cuda" in self.device.type:
            # Need to flatten data to ensure the memory is aligned following [num_vars, batch_size]
            data = data.reshape(-1).contiguous()

            tot_num_nodes = node_flows.size(0)
            batch_size = node_flows.size(1)
            node_offset = self._output_ind_range[0]

            if not self.provided("bk_local_ids"):
                layer_num_nodes = self._output_ind_range[1] - self._output_ind_range[0]
                bk_local_ids = None
            else:
                layer_num_nodes = self.bk_local_ids.size(0)
                bk_local_ids = self.bk_local_ids

            # Missing mask
            num_vars = None
            if missing_mask is None:
                missing_mask_mode = 0
            else:
                if missing_mask.dim() == 1:
                    missing_mask_mode = 1
                    num_vars = missing_mask.size(0)
                elif _batch_first or num_vars == missing_mask.size(1):
                    missing_mask_mode = 2
                    num_vars = missing_mask.size(1)
                else:
                    missing_mask_mode = 3
                    num_vars = missing_mask.size(0)

            if not self.provided("_flows_kernel"):
                self._flows_kernel = self._compile_triton_kernel(self._flows_kernel_template, flow_fn = self.bk_flow_fn)

            BLOCK_SIZE = 1024
            TILE_SIZE_K = 1

            grid = (triton.cdiv(layer_num_nodes * batch_size, BLOCK_SIZE),)

            self._flows_kernel[grid](
                params_ptr = self.params,
                param_flows_ptr = self.param_flows,
                node_flows_ptr = node_flows, 
                node_mars_ptr = node_mars,
                data_ptr = data, 
                missing_mask_ptr = missing_mask,
                vids_ptr = self.vids, 
                s_pids_ptr = self.s_pids,
                s_pfids_ptr = self.s_pfids,
                metadata_ptr = self.metadata, 
                s_mids_ptr = self.s_mids, 
                bk_local_ids_ptr = bk_local_ids,
                layer_num_nodes = layer_num_nodes, 
                batch_size = batch_size, 
                num_vars_per_node = self.num_vars_per_node, 
                num_vars = num_vars,
                nv_block_size = triton.next_power_of_2(self.num_vars_per_node),
                node_offset = node_offset, 
                BLOCK_SIZE = BLOCK_SIZE, 
                partial_eval = 1 if bk_local_ids is not None else 0,
                logspace_flows = logspace_flows,
                missing_mask_mode = missing_mask_mode,
                pass_type = 0,
                TILE_SIZE_K = 1,
                num_warps = 8
            )

            # Handle the masked input nodes
            if missing_mask is not None and self.bk_flow_mask_fn is not None:
                if not self.provided("_flows_mask_kernel"):
                    self._flows_mask_kernel = self._compile_triton_kernel(self._flows_kernel_template, flow_fn = self.bk_flow_mask_fn)

                if self._need_2nd_kernel_dim():
                    BLOCK_SIZE = 32
                    TILE_SIZE_K = 32
                else:
                    BLOCK_SIZE = 1024
                    TILE_SIZE_K = 1

                grid = (triton.cdiv(layer_num_nodes * batch_size, BLOCK_SIZE),)

                self._flows_mask_kernel[grid](
                    params_ptr = self.params,
                    param_flows_ptr = self.param_flows,
                    node_flows_ptr = node_flows, 
                    node_mars_ptr = node_mars,
                    data_ptr = data, 
                    missing_mask_ptr = missing_mask,
                    vids_ptr = self.vids, 
                    s_pids_ptr = self.s_pids,
                    s_pfids_ptr = self.s_pfids,
                    metadata_ptr = self.metadata, 
                    s_mids_ptr = self.s_mids, 
                    bk_local_ids_ptr = bk_local_ids,
                    layer_num_nodes = layer_num_nodes, 
                    batch_size = batch_size, 
                    num_vars_per_node = self.num_vars_per_node, 
                    num_vars = num_vars,
                    nv_block_size = triton.next_power_of_2(self.num_vars_per_node),
                    node_offset = node_offset, 
                    BLOCK_SIZE = BLOCK_SIZE, 
                    partial_eval = 1 if bk_local_ids is not None else 0,
                    logspace_flows = logspace_flows,
                    missing_mask_mode = missing_mask_mode,
                    pass_type = 1,
                    TILE_SIZE_K = TILE_SIZE_K,
                    num_warps = 8
                )

            # Apply post-processing kernels
            for (kernel, cond_fn, prep_kwargs_fn) in self.post_bp_fns:
                if not cond_fn(self, kwargs):
                    continue

                target_kwargs = prep_kwargs_fn(self, kwargs)

                BLOCK_SIZE = 1024

                grid = (triton.cdiv(layer_num_nodes * batch_size, BLOCK_SIZE),)

                kernel[grid](
                    params_ptr = self.params,
                    param_flows_ptr = self.param_flows,
                    node_flows_ptr = node_flows, 
                    node_mars_ptr = node_mars,
                    data_ptr = data, 
                    vids_ptr = self.vids, 
                    s_pids_ptr = self.s_pids,
                    s_pfids_ptr = self.s_pfids,
                    metadata_ptr = self.metadata, 
                    s_mids_ptr = self.s_mids, 
                    nids_ptr = self.nids,
                    bk_local_ids_ptr = bk_local_ids,
                    layer_num_nodes = layer_num_nodes, 
                    batch_size = batch_size, 
                    num_vars_per_node = self.num_vars_per_node, 
                    num_vars = num_vars,
                    nv_block_size = triton.next_power_of_2(self.num_vars_per_node),
                    node_offset = node_offset, 
                    BLOCK_SIZE = BLOCK_SIZE, 
                    partial_eval = 1 if bk_local_ids is not None else 0,
                    logspace_flows = logspace_flows,
                    TILE_SIZE_K = 1,
                    num_warps = 8,
                    **target_kwargs
                )
        else:
            raise NotImplementedError("CPU backward fn for input nodes is not implemented.")

    def sample(self, samples: torch.Tensor, node_flows: torch.Tensor, missing_mask: Optional[torch.Tensor] = None, 
               params: Optional[torch.Tensor] = None, seed: Optional[int] = None):
        """
        samples:       [num_vars, B]
        missing_mask:  [num_vars, B] or [num_vars] or None
        node_flows:    [num_nodes, B]
        """

        if params is None:
            params = self.params
        else:
            params = params["params"]

        assert params.dim() == 1

        if "cuda" in self.device.type:

            sid, eid = self._output_ind_range
            tot_num_nodes = node_flows.size(0)
            batch_size = node_flows.size(1)
            node_offset = self._output_ind_range[0]

            # Get all node ids with non-zero flow
            nflow_xids, nflow_yids = torch.where(node_flows[sid:eid,:] > 1e-8)
            num_activ_nodes = nflow_xids.size(0)

            if not self.provided("_sample_kernel"):
                self._sample_kernel = self._compile_triton_kernel(self._sample_kernel_template, sample_fn = self.sample_fn)

            BLOCK_SIZE = 1024

            grid = (triton.cdiv(num_activ_nodes, BLOCK_SIZE),)

            self._sample_kernel[grid](
                samples_ptr = samples, 
                params_ptr = params,
                nflow_xids_ptr = nflow_xids, 
                nflow_yids_ptr = nflow_yids, 
                vids_ptr = self.vids, 
                s_pids_ptr = self.s_pids, 
                metadata_ptr = self.metadata,
                s_mids_ptr = self.s_mids,
                num_activ_nodes = num_activ_nodes, 
                num_vars_per_node = self.num_vars_per_node, 
                nv_block_size = triton.next_power_of_2(self.num_vars_per_node),
                batch_size = batch_size, 
                BLOCK_SIZE = BLOCK_SIZE,
                seed = seed if seed is not None else random.randint(0, int(1e8))
            )

        else:
            raise NotImplementedError("CPU sample fn for input nodes is not implemented.")

    def mini_batch_em(self, step_size: float, pseudocount: float = 0.0):
        if not self._used_external_params:
            # Normalize and update parameters
            with torch.no_grad():

                if "cuda" in self.device.type:

                    # Accumulate parameter flows of tied nodes
                    for i in range(len(self.tied2source_nids)):
                        pfid_start, num_par_flows, ch_pfids = self.tied2source_nids[i]
                        num_coalesced_blocks = ch_pfids.size(0)

                        if num_coalesced_blocks <= 1024:
                            BLOCK_N = triton.next_power_of_2(num_coalesced_blocks)
                            BLOCK_M = min(1024 // BLOCK_N, num_par_flows)

                            grid = (triton.cdiv(num_par_flows, BLOCK_M),)

                            self._pflow_accum_kernel[grid](
                                param_flows_ptr = self.param_flows,
                                pfid_start = pfid_start,
                                ch_pfids_ptr = ch_pfids,
                                num_coalesced_blocks = num_coalesced_blocks,
                                num_par_flows = num_par_flows,
                                BLOCK_M = BLOCK_M,
                                BLOCK_N = BLOCK_N
                            )
                        else:
                            raise NotImplementedError("Unsupported number of coalesced parameter flows.")


                    layer_num_source_nodes = self.source_nids.size(0)

                    if not self.provided("_em_kernel"):
                        self._em_kernel = self._compile_triton_kernel(self._em_kernel_template, em_fn = self.em_fn)

                    constexprs = torch.tensor([step_size, pseudocount], dtype = torch.float32, device = self.device)

                    if hasattr(self.nodes[0].dist, "em_block_size"):
                        BLOCK_SIZE = self.nodes[0].dist.em_block_size
                    else:
                        BLOCK_SIZE = 1024

                    grid = (triton.cdiv(layer_num_source_nodes, BLOCK_SIZE),)

                    self._em_kernel[grid](
                        params_ptr = self.params,
                        param_flows_ptr = self.param_flows,
                        s_pids_ptr = self.s_pids,
                        s_pfids_ptr = self.s_pfids,
                        metadata_ptr = self.metadata,
                        s_mids_ptr = self.s_mids,
                        source_nids_ptr = self.source_nids,
                        constexprs_ptr = constexprs,
                        layer_num_source_nodes = layer_num_source_nodes,
                        BLOCK_SIZE = BLOCK_SIZE,
                        num_warps = 8
                    )

                else:
                    raise NotImplementedError("CPU minibatch em fn for input nodes is not implemented.")

    def eval_partition_fn(self, node_mars: torch.Tensor, params: Optional[Dict] = None):
        """
        Evaluate the partition function.
        node_mars: [num_nodes]
        """

        if params is None:
            params = self.params
        else:
            params = params["params"]

        assert params.dim() == 1

        if "cuda" in self.device.type:
            node_offset = self._output_ind_range[0]
            layer_num_nodes = self._output_ind_range[1] - self._output_ind_range[0]

            if not self.provided("_partition_fn_kernel"):
                self._partition_fn_kernel = self._compile_triton_kernel(self._partition_fn_kernel_template, partition_fn = self.fw_partition_fn)

            if self._need_2nd_kernel_dim():
                BLOCK_SIZE = 32
                TILE_SIZE_K = 32
            else:
                BLOCK_SIZE = 1024
                TILE_SIZE_K = 1

            grid = (triton.cdiv(layer_num_nodes, BLOCK_SIZE),)

            self._partition_fn_kernel[grid](
                params_ptr = self.params,
                node_mars_ptr = node_mars,
                s_pids_ptr = self.s_pids,
                metadata_ptr = self.metadata,
                s_mids_ptr = self.s_mids,
                layer_num_nodes = layer_num_nodes,
                node_offset = node_offset,
                BLOCK_SIZE = BLOCK_SIZE,
                TILE_SIZE_K = TILE_SIZE_K,
                num_warps = 8
            )

        else:
            raise NotImplementedError("CPU minibatch partition fn for input nodes is not implemented.")

    def add_missing_flows(self, node_flows: torch.Tensor, node_mars: Optional[torch.Tensor] = None, logspace_flows: bool = False, 
                          scale: float = 1.0, pc_is_normalized: bool = True, _pre_accum_nflows: bool = True):
        """
        Add missing flows specified by `node_flows` to the input node parameters.
        node_flows: [num_nodes]
        """

        if not pc_is_normalized:
            sid, eid = self._output_ind_range
            node_flows[sid:eid] /= torch.exp(node_mars[sid:eid])

        if "cuda" in self.device.type:
            node_offset = self._output_ind_range[0]
            layer_num_nodes = self._output_ind_range[1] - self._output_ind_range[0]

            if not self.provided("_missing_flows_kernel"):
                assert self.bk_flow_mask_fn is not None, "The target distribution doesn't have `bk_flow_mask_fn` implemented."
                self._missing_flows_kernel = self._compile_triton_kernel(self._missing_flows_kernel_template, flow_fn = self.bk_flow_mask_fn)

            if self._need_2nd_kernel_dim():
                BLOCK_SIZE = 32
                TILE_SIZE_K = 32
            else:
                BLOCK_SIZE = 1024
                TILE_SIZE_K = 1

            if _pre_accum_nflows:
                layer_num_source_nodes = self.source_nids.size(0)

                self._accum_flows_to_tied_nodes(node_flows, logspace_flows)

                grid = (triton.cdiv(layer_num_source_nodes, BLOCK_SIZE),)

                self._missing_flows_kernel[grid](
                    node_flows_ptr = node_flows,
                    params_ptr = self.params,
                    param_flows_ptr = self.param_flows,
                    s_pids_ptr = self.s_pids,
                    s_pfids_ptr = self.s_pfids,
                    metadata_ptr = self.metadata, 
                    s_mids_ptr = self.s_mids, 
                    source_nids_ptr = self.source_nids,
                    scale = scale,
                    node_offset = node_offset,
                    layer_num_nodes = layer_num_nodes,
                    logspace_flows = logspace_flows,
                    are_source_nodes = True,
                    BLOCK_SIZE = BLOCK_SIZE,
                    TILE_SIZE_K = TILE_SIZE_K
                )

            else:
                grid = (triton.cdiv(layer_num_nodes, BLOCK_SIZE),)

                self._missing_flows_kernel[grid](
                    node_flows_ptr = node_flows,
                    params_ptr = self.params,
                    param_flows_ptr = self.param_flows,
                    s_pids_ptr = self.s_pids,
                    s_pfids_ptr = self.s_pfids,
                    metadata_ptr = self.metadata, 
                    s_mids_ptr = self.s_mids, 
                    source_nids_ptr = self.source_nids,
                    scale = scale,
                    node_offset = node_offset,
                    layer_num_nodes = layer_num_nodes,
                    logspace_flows = logspace_flows,
                    are_source_nodes = False,
                    BLOCK_SIZE = BLOCK_SIZE,
                    TILE_SIZE_K = TILE_SIZE_K
                )
        else:
            raise NotImplementedError("CPU minibatch missing flow fn for input nodes is not implemented.")

    def _accum_flows_to_tied_nodes(self, node_flows: torch.Tensor, logspace_flows: bool = False):

        assert not logspace_flows, "Not implemented yet."

        if "cuda" in self.device.type:

            node_offset = self._output_ind_range[0]
            layer_num_nodes = self._output_ind_range[1] - self._output_ind_range[0]
            
            BLOCK_SIZE = 1024

            grid = (triton.cdiv(layer_num_nodes, BLOCK_SIZE),)

            self._accum_flows_to_tied_nodes_kernel[grid](
                node_flows_ptr = node_flows,
                node_id2source_id_ptr = self.node_id2source_id,
                node_offset = node_offset,
                layer_num_nodes = layer_num_nodes,
                BLOCK_SIZE = BLOCK_SIZE
            )

        else:
            raise NotImplementedError("CPU accum flow fn for input nodes is not implemented.")

    def get_param_specs(self):
        return {"params": torch.Size([self.num_parameters])}

    def get_data_dtype(self):
        """
        Get the data dtype for the distribution.
        """
        return self.nodes[0].get_data_dtype()

    def enable_partial_evaluation(self, fw_scopes: Optional[Union[Sequence[BitSet],Sequence[int]]] = None, 
                                  bk_scopes: Optional[Union[Sequence[BitSet],Sequence[int]]] = None, return_ids: bool = False):
        # Create cache if needed
        if not self.provided("scope2localgids"):
            self._prepare_scope2nids()

        # Filter forward nodes
        if fw_scopes is not None:
            fw_local_ids = []
            for scope in fw_scopes:
                if isinstance(scope, int):
                    scope = BitSet.from_array([scope])

                if scope not in self.scope2localgids:
                    continue

                fw_local_ids.append(self.scope2localgids[scope])

            if return_ids:
                return torch.cat(fw_local_ids, dim = 0)
            else:
                self.fw_local_ids = torch.cat(fw_local_ids, dim = 0)

        # Filter backward nodes
        if bk_scopes is not None:
            bk_local_ids = []
            for scope in bk_scopes:
                if isinstance(scope, int):
                    scope = BitSet.from_array([scope])

                if scope not in self.scope2localgids:
                    continue

                bk_local_ids.append(self.scope2localgids[scope])

            if return_ids:
                return torch.cat(bk_local_ids, dim = 0)
            else:
                self.bk_local_ids = torch.cat(bk_local_ids, dim = 0)

    def disable_partial_evaluation(self, forward: bool = True, backward: bool = True):
        if forward:
            self.fw_local_ids = None

        if backward:
            self.bk_local_ids = None

    def update_parameters(self):
        for idx, ns in enumerate(self.nodes):
            if ns.is_tied():
                continue

            par_start, par_end = ns._param_range
            ns._params = self.params.data[par_start:par_end].detach().cpu().clone()

    def _prepare_scope2nids(self):
        if not hasattr(self, "scope2localgids"):
            scope2localgids = dict()

            local_ngid = 0
            for ns in self.nodes:
                scope = ns.scope

                s_ngid = local_ngid
                e_ngid = local_ngid + ns.num_node_blocks

                with torch.no_grad():
                    if scope not in scope2localgids:
                        scope2localgids[scope] = [torch.zeros([0], dtype = torch.long)]

                    scope2localgids[scope].append(torch.arange(s_ngid, e_ngid))

                local_ngid += ns.num_node_blocks

            self.scope2localgids = {
                scope: torch.cat(ids, dim = 0).to(self.params.device) for scope, ids in scope2localgids.items()
            }

    def is_input(self):
        return True

    def _reorder_nodes(self, nodes):
        node_set = set(nodes)
        reordered_untied_nodes = []
        reordered_tied_nodes = []
        added_node_set = set()
        for ns in nodes:
            if ns in added_node_set:
                continue
            if not ns.is_tied():
                reordered_untied_nodes.append(ns)
                added_node_set.add(ns)
            else:
                source_ns = ns.get_source_ns()
                if source_ns in added_node_set:
                    reordered_tied_nodes.append(ns)
                    added_node_set.add(ns)
                elif source_ns in node_set:
                    reordered_untied_nodes.append(source_ns)
                    reordered_tied_nodes.append(ns)
                    added_node_set.add(ns)
                    added_node_set.add(source_ns)
                else:
                    raise ValueError("A tied `InputNodes` should be in the same layer with its source nodes.")

        reordered_nodes = reordered_untied_nodes + reordered_tied_nodes

        assert len(reordered_nodes) == len(nodes), "Total node length should not change after reordering."

        return reordered_nodes

    def _init_parameters(self, perturbation):

        p_start, p_end = 0, 0
        for ns_id, ns in enumerate(self.nodes):
            # `params` (init/copy parameters)
            if not ns.is_tied():
                p_end = p_start + ns.num_nodes * ns.dist.num_parameters()
                if ns.has_params():
                    self.params[p_start:p_end] = ns._params.to(self.device)
                else:
                    self.params[p_start:p_end] = ns.init_parameters(ret_params = True).to(self.device)

                p_start = p_end

    def _need_2nd_kernel_dim(self):
        return self.nodes[0].dist._need_2nd_kernel_dim()

    @staticmethod
    def _mars_kernel_template(mar_fn, params_ptr, node_mars_ptr, data_ptr, vids_ptr, s_pids_ptr, metadata_ptr, s_mids_ptr, 
                              fw_local_ids_ptr, partial_eval: tl.constexpr, layer_num_nodes: tl.constexpr, batch_size: tl.constexpr, 
                              num_vars_per_node: tl.constexpr, nv_block_size: tl.constexpr, node_offset: tl.constexpr, BLOCK_SIZE: tl.constexpr):
        pid = tl.program_id(axis = 0)
        block_start = pid * BLOCK_SIZE

        offsets = block_start + tl.arange(0, BLOCK_SIZE)
        mask = offsets < layer_num_nodes * batch_size

        # Raw batch and (local) node id
        batch_offsets = (offsets % batch_size)
        local_offsets = (offsets // batch_size)

        if partial_eval > 0:
            local_offsets = tl.load(fw_local_ids_ptr + local_offsets, mask = mask, other = 0)

        if num_vars_per_node == 1:
            # Get all variable ids
            vids = tl.load(vids_ptr + local_offsets, mask = mask, other = 0)

            # Load the corresponding data
            data = tl.load(data_ptr + vids * batch_size + batch_offsets, mask = mask, other = 0)
        else:
            # Get all variable ids
            vids_offsets = tl.broadcast_to(local_offsets[:,None], (BLOCK_SIZE, nv_block_size)) * num_vars_per_node + \
                tl.broadcast_to(tl.arange(0, nv_block_size)[None,:], (BLOCK_SIZE, nv_block_size))
            vids_mask = tl.broadcast_to(mask[:,None], (BLOCK_SIZE, nv_block_size)) & \
                tl.broadcast_to((tl.arange(0, nv_block_size) < num_vars_per_node)[None,:], (BLOCK_SIZE, nv_block_size))
            vids = tl.load(vids_ptr + vids_offsets, mask = vids_mask, other = 0)

            # Load the corresponding data
            data = tl.load(data_ptr + vids * batch_size + batch_offsets, mask = vids_mask, other = 0)

        s_pids = tl.load(s_pids_ptr + local_offsets, mask = mask, other = 0)

        mars = mar_fn(local_offsets, data, params_ptr, s_pids, metadata_ptr, s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE)

        node_offsets = local_offsets + node_offset
        tl.store(node_mars_ptr + node_offsets * batch_size + batch_offsets, mars, mask = mask)

    @staticmethod
    # @triton.jit
    @triton_jit
    def _fw_missing_mask_kernel(missing_mask_ptr, node_mars_ptr, vids_ptr, fw_local_ids_ptr, num_vars,
                                layer_num_nodes: tl.constexpr, batch_size: tl.constexpr, node_offset: tl.constexpr, 
                                BLOCK_SIZE: tl.constexpr, partial_eval: tl.constexpr, mode: tl.constexpr):
        pid = tl.program_id(axis = 0)
        block_start = pid * BLOCK_SIZE

        offsets = block_start + tl.arange(0, BLOCK_SIZE)
        mask = offsets < layer_num_nodes * batch_size

        # Raw batch and (local) node id
        batch_offsets = (offsets % batch_size)
        local_offsets = (offsets // batch_size)

        if partial_eval > 0:
            local_offsets = tl.load(fw_local_ids_ptr + local_offsets, mask = mask, other = 0)

        # Get all variable ids
        vids = tl.load(vids_ptr + local_offsets, mask = mask, other = 0)

        # Fetch mask
        if mode == 0:
            # `mask_dim == 1`
            missing_mask = tl.load(missing_mask_ptr + vids, mask = mask, other = False)
        elif mode == 1:
            # `mask_dim == 1` and first dimension is batch
            mask_offsets = vids + batch_offsets * num_vars
            missing_mask = tl.load(missing_mask_ptr + mask_offsets, mask = mask, other = False)
        elif mode == 2:
            # `mask_dim == 1` and second dimension is batch
            mask_offsets = vids * batch_size + batch_offsets
            missing_mask = tl.load(missing_mask_ptr + mask_offsets, mask = mask, other = False)

        # Apply mask
        node_offsets = (local_offsets + node_offset) * batch_size + batch_offsets
        mars = tl.load(node_mars_ptr + node_offsets, mask = mask, other = 0.0)
        mars = tl.where(missing_mask, 0.0, mars)
        tl.store(node_mars_ptr + node_offsets, mars, mask = mask)

    @staticmethod
    def _flows_kernel_template(flow_fn, params_ptr, param_flows_ptr, node_flows_ptr, node_mars_ptr, data_ptr, missing_mask_ptr, vids_ptr, s_pids_ptr, s_pfids_ptr,
                               metadata_ptr, s_mids_ptr, bk_local_ids_ptr, partial_eval: tl.constexpr, logspace_flows: tl.constexpr, layer_num_nodes: tl.constexpr, 
                               batch_size: tl.constexpr, num_vars_per_node: tl.constexpr, num_vars: tl.constexpr, nv_block_size: tl.constexpr, node_offset: tl.constexpr, 
                               missing_mask_mode: tl.constexpr, pass_type: tl.constexpr, BLOCK_SIZE: tl.constexpr, TILE_SIZE_K: tl.constexpr):
        pid = tl.program_id(axis = 0)
        block_start = pid * BLOCK_SIZE

        offsets = block_start + tl.arange(0, BLOCK_SIZE)
        mask = offsets < layer_num_nodes * batch_size

        # Raw batch and (local) node id
        batch_offsets = (offsets % batch_size)
        local_offsets = (offsets // batch_size)

        if partial_eval > 0:
            local_offsets = tl.load(bk_local_ids_ptr + local_offsets, mask = mask, other = 0)

        if num_vars_per_node == 1:
            # Get all variable ids
            vids = tl.load(vids_ptr + local_offsets, mask = mask, other = 0)

            # Load the corresponding data
            data = tl.load(data_ptr + vids * batch_size + batch_offsets, mask = mask, other = 0)

            if missing_mask_mode == 1:
                missing_mask = tl.load(missing_mask_ptr + vids, mask = mask, other = False).to(tl.int1)
                if pass_type == 0:
                    mask &= (missing_mask == False)
            elif missing_mask_mode == 2:
                missing_mask = tl.load(missing_mask_ptr + batch_offsets * num_vars + vids, mask = mask, other = False).to(tl.int1)
                if pass_type == 0:
                    mask &= (missing_mask == False)
            elif missing_mask_mode == 3:
                missing_mask = tl.load(missing_mask_ptr + vids * batch_size + batch_offsets, mask = mask, other = False).to(tl.int1)
                if pass_type == 0:
                    mask &= (missing_mask == False)
        else:
            # Get all variable ids
            vids_offsets = tl.broadcast_to(local_offsets[:,None], (BLOCK_SIZE, nv_block_size)) * num_vars_per_node + \
                tl.broadcast_to(tl.arange(0, nv_block_size)[None,:], (BLOCK_SIZE, nv_block_size))
            vids_mask = tl.broadcast_to(mask[:,None], (BLOCK_SIZE, nv_block_size)) & \
                tl.broadcast_to((tl.arange(0, nv_block_size) < num_vars_per_node)[None,:], (BLOCK_SIZE, nv_block_size))
            vids = tl.load(vids_ptr + vids_offsets, mask = vids_mask, other = 0)

            # Load the corresponding data
            data = tl.load(data_ptr + vids * batch_size + batch_offsets, mask = vids_mask, other = 0)

        s_pids = tl.load(s_pids_ptr + local_offsets, mask = mask, other = 0)
        s_pfids = tl.load(s_pfids_ptr + local_offsets, mask = mask, other = 0)

        ns_offsets = (local_offsets + node_offset) * batch_size + batch_offsets
        flows = tl.load(node_flows_ptr + ns_offsets, mask = mask, other = 0)

        if logspace_flows:
            flows = tl.exp(flows)

        flow_fn(local_offsets, ns_offsets, data, flows, node_mars_ptr, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, 
                s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE)

    @staticmethod
    def _missing_flows_kernel_template(node_flows_ptr, params_ptr, param_flows_ptr, s_pids_ptr, s_pfids_ptr,
                                       metadata_ptr, s_mids_ptr, source_nids_ptr, scale, node_offset: tl.constexpr, layer_num_nodes: tl.constexpr, 
                                       logspace_flows: tl.constexpr, are_source_nodes: tl.constexpr, BLOCK_SIZE: tl.constexpr, TILE_SIZE_K: tl.constexpr):
        pid = tl.program_id(axis = 0)
        block_start = pid * BLOCK_SIZE

        offsets = block_start + tl.arange(0, BLOCK_SIZE)
        mask = offsets < layer_num_nodes

        # Get the local node ids
        if are_source_nodes:
            local_offsets = tl.load(source_nids_ptr + offsets, mask = mask, other = 0)
        else:
            local_offsets = offsets

        s_pids = tl.load(s_pids_ptr + local_offsets, mask = mask, other = 0)
        s_pfids = tl.load(s_pfids_ptr + local_offsets, mask = mask, other = 0)

        ns_offsets = local_offsets + node_offset
        flows = tl.load(node_flows_ptr + ns_offsets, mask = mask, other = 0)

        if logspace_flows:
            flows = tl.exp(flows)

        flows *= scale

        missing_mask = True

        flow_fn(local_offsets, ns_offsets, data, flows, node_mars_ptr, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, 
                s_mids_ptr, mask, num_vars_per_node, BLOCK_SIZE)

    @staticmethod
    @triton_jit
    def _accum_flows_to_tied_nodes_kernel(node_flows_ptr, node_id2source_id_ptr, node_offset, layer_num_nodes, BLOCK_SIZE: tl.constexpr):
        pid = tl.program_id(axis = 0)
        block_start = pid * BLOCK_SIZE

        offsets = block_start + tl.arange(0, BLOCK_SIZE)
        mask = offsets < layer_num_nodes

        source_ids = tl.load(node_id2source_id_ptr + offsets, mask = mask, other = -1)
        mask = mask & (source_ids != -1)

        flows = tl.load(node_flows_ptr + node_offset + offsets, mask = mask, other = 0)

        tl.atomic_add(node_flows_ptr + node_offset + source_ids, flows, mask = mask)

    @staticmethod
    def _partition_fn_kernel_template(partition_fn, params_ptr, node_mars_ptr, s_pids_ptr, metadata_ptr, s_mids_ptr, layer_num_nodes: tl.constexpr,
                                      node_offset: tl.constexpr, BLOCK_SIZE: tl.constexpr, TILE_SIZE_K: tl.constexpr):
        pid = tl.program_id(axis = 0)
        block_start = pid * BLOCK_SIZE

        offsets = block_start + tl.arange(0, BLOCK_SIZE)
        mask = offsets < layer_num_nodes

        s_pids = tl.load(s_pids_ptr + offsets, mask = mask, other = 0)

        mars = partition_fn(offsets, params_ptr, s_pids, metadata_ptr, s_mids_ptr, mask, BLOCK_SIZE, TILE_SIZE_K)

        node_offsets = node_offset + offsets
        tl.store(node_mars_ptr + node_offsets, mars, mask = mask)

    @staticmethod
    def _sample_kernel_template(sample_fn, samples_ptr, params_ptr, nflow_xids_ptr, nflow_yids_ptr, vids_ptr, s_pids_ptr, metadata_ptr, s_mids_ptr,
                                seed, num_vars_per_node: tl.constexpr, nv_block_size: tl.constexpr, num_activ_nodes: tl.constexpr, 
                                batch_size: tl.constexpr, BLOCK_SIZE: tl.constexpr):
        pid = tl.program_id(axis = 0)
        block_start = pid * BLOCK_SIZE

        offsets = block_start + tl.arange(0, BLOCK_SIZE)
        mask = offsets < num_activ_nodes

        # Raw batch and (local) node id
        local_offsets = tl.load(nflow_xids_ptr + offsets, mask = mask, other = 0)
        batch_offsets = tl.load(nflow_yids_ptr + offsets, mask = mask, other = 0)

        # Get all variable ids
        if num_vars_per_node == 1:
            vids = tl.load(vids_ptr + local_offsets, mask = mask, other = 0)
        else:
            # Get all variable ids
            vids_offsets = tl.broadcast_to(local_offsets[:,None], (BLOCK_SIZE, nv_block_size)) * num_vars_per_node + \
                tl.broadcast_to(tl.arange(0, nv_block_size)[None,:], (BLOCK_SIZE, nv_block_size))
            vids_mask = tl.broadcast_to(mask[:,None], (BLOCK_SIZE, nv_block_size)) & \
                tl.broadcast_to((tl.arange(0, nv_block_size) < num_vars_per_node)[None,:], (BLOCK_SIZE, nv_block_size))
            vids = tl.load(vids_ptr + vids_offsets, mask = vids_mask, other = 0)

        # Get start parameter indices
        s_pids = tl.load(s_pids_ptr + local_offsets, mask = mask, other = 0)

        sample_fn(samples_ptr, local_offsets, batch_offsets, vids, s_pids, params_ptr, metadata_ptr, s_mids_ptr, mask, batch_size, BLOCK_SIZE, seed)

    @staticmethod
    # @triton.jit
    @triton_jit
    def _pflow_accum_kernel(param_flows_ptr, pfid_start, ch_pfids_ptr, num_coalesced_blocks, num_par_flows, BLOCK_M: tl.constexpr, BLOCK_N: tl.constexpr):
        pid = tl.program_id(axis = 0)

        offs_pflow = pid * BLOCK_M + tl.arange(0, BLOCK_M)
        mask_pflow = offs_pflow < num_par_flows

        offs_ch = tl.arange(0, BLOCK_N)
        mask_ch = offs_ch < num_coalesced_blocks

        # Start id for all ch parflows
        ch_pstart = tl.load(ch_pfids_ptr + offs_ch, mask = mask_ch)

        offs_ch_pflow = offs_pflow[:,None] + ch_pstart[None,:]
        mask_ch_pflow = mask_pflow[:,None] & mask_ch[None,:]
        ch_pflows = tl.load(param_flows_ptr + offs_ch_pflow, mask = mask_ch_pflow, other = 0)

        tar_pflows = tl.sum(ch_pflows, axis = 1)

        tl.store(param_flows_ptr + pfid_start + offs_pflow, tar_pflows, mask = mask_pflow)

    @staticmethod
    def _em_kernel_template(em_fn, params_ptr, param_flows_ptr, s_pids_ptr, s_pfids_ptr, metadata_ptr, s_mids_ptr,
                            source_nids_ptr, constexprs_ptr, layer_num_source_nodes: tl.constexpr, BLOCK_SIZE: tl.constexpr):
        pid = tl.program_id(axis = 0)
        block_start = pid * BLOCK_SIZE

        # Retrieve all constexprs
        step_size = tl.load(constexprs_ptr)
        pseudocount = tl.load(constexprs_ptr + 1)

        offsets = block_start + tl.arange(0, BLOCK_SIZE)
        mask = offsets < layer_num_source_nodes

        # Get the local node ids
        local_offsets = tl.load(source_nids_ptr + offsets, mask = mask, other = 0)

        # Get the corresponding start id for `params` and `param_flows`
        s_pids = tl.load(s_pids_ptr + local_offsets, mask = mask, other = 0)
        s_pfids = tl.load(s_pfids_ptr + local_offsets, mask = mask, other = 0)

        em_fn(local_offsets, params_ptr, param_flows_ptr, s_pids, s_pfids, metadata_ptr, s_mids_ptr, mask,
              step_size, pseudocount, BLOCK_SIZE)

    @staticmethod
    def _compile_triton_kernel(main_fn, **sub_fns):

        def parse_source(src, get_signature = False):

            if not isinstance(src, str):
                src = textwrap.dedent(inspect.getsource(src))
                src = src[src.find("def"):]

            fn_header = ""
            fn_body = []
            curr_body = ""
            is_header = True
            for line in src.split("\n"):
                if is_header:
                    fn_header += line.strip(" ")
                    if ")" in line:
                        is_header = False
                else:
                    if line.strip(" ").startswith("#") or len(line.strip(" ")) == 0:
                        continue

                    if len(curr_body) > 0:
                        line = line.strip(" ")
                    else:
                        line = line.rstrip(" ")
                    curr_body += line
                    if curr_body[-1] == "\\":
                        curr_body = curr_body[:-1]
                    else:
                        parenthesis_count = curr_body.count("(") - curr_body.count(")")
                        if parenthesis_count == 0:
                            fn_body.append(curr_body)
                            curr_body = ""
                            parenthesis_count = 0

            if get_signature:
                args = fn_header.split("(")[1].split(")")[0].split(",")
                signature = list(map(lambda arg: arg.split(":")[0].split("=")[0].strip(" "), args))

                return fn_header, fn_body, signature

            return fn_header, fn_body

        main_fn_header, main_fn_body = parse_source(main_fn)

        # Map to new function header
        fn_name = main_fn_header.split("(")[0]
        fn_args = main_fn_header.split("(")[1].split(")")[0].split(",")
        fn_args = list(map(
            lambda str: str.strip(" "),
            filter(lambda arg: arg.split(":")[0].strip(" ") not in sub_fns, fn_args)
        ))
        seed_str = f"_{random.randint(0,int(1e8))}"
        new_fn_header = fn_name + seed_str + "(" + ",".join(fn_args) + "):"
        global_key = fn_name.split("def")[1].strip(" ") + seed_str

        # Map to new function body
        new_fn_body = []
        sub_fns = {k: parse_source(v, get_signature = True) for k, v in sub_fns.items()}
        for line in main_fn_body:
            substituted = False
            for sub_fn_name in sub_fns:
                if sub_fn_name in line:

                    # Get the target variable name
                    target_var = line.split(sub_fn_name)[0].split("=")[0].strip(" ")

                    # Get the variable name mapping
                    target_args = line.split(sub_fn_name)[1].split("(")[1].split(")")[0].split(",")
                    var_mapping = {k: v.strip(" ") for k, v in zip(sub_fns[sub_fn_name][2], target_args)}

                    indent = " " * (len(line) - len(line.lstrip()))
                    base_indent_len = len(sub_fns[sub_fn_name][1][0]) - len(sub_fns[sub_fn_name][1][0].lstrip())
                    for fn_line in sub_fns[sub_fn_name][1]:
                        fn_line = indent + fn_line[base_indent_len:]
                        if fn_line.split("#")[0].strip(" ") == "pass":
                            continue
                        for k, v in var_mapping.items():
                            fn_line = fn_line.replace(k, v)
                        if "return" in fn_line and len(target_var) > 0:
                            fn_line = fn_line.replace("return", f"{target_var} =")
                        new_fn_body.append(fn_line)

                    substituted = True
                    break

            if not substituted:
                new_fn_body.append(line)

        # Updated source code
        new_src = new_fn_header + "\n" + "\n".join(new_fn_body)

        # Add import commands
        new_src = "import triton\nimport triton.language as tl\nif hasattr(tl.extra.cuda, 'libdevice'):\n    tlmath = tl.extra.cuda.libdevice\nelse:\n    tlmath = tl.math\n\n" + new_src

        # Make a pseudo-function from the source code
        new_fn = make_function_from_src(new_src)

        return triton_jit(new_fn)



from __future__ import annotations

import torch
import numpy as np
import networkx as nx
from typing import Type, Optional

from pyjuice.nodes.distributions import *
from .compilation import BayesianTreeToHiddenRegionGraph


def mutual_information(x1: torch.Tensor, x2: torch.Tensor, num_bins: int, sigma: float):
    assert x1.device == x2.device

    device = x1.device
    B, K1 = x1.size()
    K2 = x2.size(1)

    x1 = (x1 - torch.min(x1)) / (torch.max(x1) - torch.min(x1) + 1e-8)
    x2 = (x2 - torch.min(x2)) / (torch.max(x2) - torch.min(x2) + 1e-8)

    bins = torch.linspace(0, 1, num_bins, device = device)

    x1p = torch.exp(-0.5 * (x1.unsqueeze(2) - bins.view(1, 1, -1)).pow(2) / sigma**2) # (B, K1, n_bin)
    x2p = torch.exp(-0.5 * (x2.unsqueeze(2) - bins.view(1, 1, -1)).pow(2) / sigma**2) # (B, K2, n_bin)

    x12p = torch.einsum("bia,baj->ij", x1p.reshape(B, K1 * num_bins, 1), x2p.reshape(B, 1, K2 * num_bins)).reshape(K1, num_bins, K2, num_bins) / B

    x1p_norm = (x1p / x1p.sum(dim = 2, keepdim = True)).mean(dim = 0)
    x2p_norm = (x2p / x2p.sum(dim = 2, keepdim = True)).mean(dim = 0)
    x12p_norm = x12p / x12p.sum(dim = (1, 3), keepdim = True) # (K1, n_bin, K2, n_bin)

    m1 = -(x1p_norm * torch.log(x1p_norm + 1e-4)).sum(dim = 1)
    m2 = -(x2p_norm * torch.log(x2p_norm + 1e-4)).sum(dim = 1)
    m12 = -(x12p_norm * torch.log(x12p_norm + 1e-4)).sum(dim = (1, 3))

    mi = m1.unsqueeze(1) + m2.unsqueeze(0) - m12
    return mi


def mutual_information_chunked(x1: torch.Tensor, x2: torch.Tensor, num_bins: int, sigma: float, chunk_size: int):
    K = x1.size(1)
    mi = torch.zeros([K, K])
    for x_s in range(0, K, chunk_size):
        x_e = min(x_s + chunk_size, K)
        for y_s in range(0, K, chunk_size):
            y_e = min(y_s + chunk_size, K)

            mi[x_s:x_e,y_s:y_e] = mutual_information(x1[:,x_s:x_e], x2[:,y_s:y_e], num_bins, sigma)

    return mi


def chow_liu_tree(mi: np.ndarray):
    K = mi.shape[0]
    G = nx.Graph()
    for v in range(K):
        G.add_node(v)
        for u in range(v):
            G.add_edge(u, v, weight = -mi[u, v])

    T = nx.minimum_spanning_tree(G)

    return T
    

def HCLT(x: torch.Tensor, num_latents: int, 
         num_bins: int = 32, 
         sigma: float = 0.5 / 32,
         chunk_size: int = 64,
         num_root_ns: int = 1,
         block_size: Optional[int] = None,
         input_dist: Optional[Distribution] = None,
         input_node_type: Type[Distribution] = Categorical, 
         input_node_params: dict = {"num_cats": 256},
         tie_input_params: bool = False):
    """
    Construct Hidden Chow-Liu Trees (https://arxiv.org/pdf/2106.02264.pdf).

    :param x: the input data of size [# samples, # variables] used to construct the backbone Chow-Liu Tree
    :type x: torch.Tensor

    :param num_latents: size of the latent space
    :type num_latents: int

    :param num_bins: number of bins to divide the input data for mutual information estimation
    :type num_bins: int

    :param sigma: a variation parameter used when estimating mutual information
    :type sigma: float

    :param chunk_size: chunk size to compute mutual information (consider decreasing if running out of GPU memory)
    :type chunk_size: int

    :param num_root_ns: number of root nodes
    :type num_root_ns: int

    :param block_size: block size
    :type block_size: int

    :param input_dist: input distribution
    :type input_dist: Distribution

    :param tie_input_params: whether to tie the input parameters
    :type tie_input_params: bool
    """

    if input_dist is not None:
        input_node_type, input_node_params = input_dist._get_constructor()
    
    mi = mutual_information_chunked(x, x, num_bins, sigma, chunk_size = chunk_size).detach().cpu().numpy()
    T = chow_liu_tree(mi)
    root = nx.center(T)[0]

    root_r = BayesianTreeToHiddenRegionGraph(
        T, root, num_latents, input_node_type, 
        input_node_params, num_root_ns = num_root_ns,
        block_size = block_size, tie_input_params = tie_input_params
    )
    
    return root_r


"""
Credit to Anthony Sottile: 
https://stackoverflow.com/questions/64925104/inspect-getsource-from-a-function-defined-in-a-string-s-def-f-return-5
"""

import os.path
import sys
import tempfile
from importlib.util import module_from_spec, spec_from_loader
from types import ModuleType
from typing import Any, Callable


class ShowSourceLoader:
    def __init__(self, modname: str, source: str) -> None:
        self.modname = modname
        self.source = source

    def get_source(self, modname: str) -> str:
        if modname != self.modname:
            raise ImportError(modname)
        return self.source


def make_function_from_src(s: str) -> Callable[..., Any]:
    filename = tempfile.mktemp(suffix = '.py')
    modname = os.path.splitext(os.path.basename(filename))[0]
    assert modname not in sys.modules

    # Get function name
    fn_name = s.split("def")[1].split("(")[0].strip(" ")

    # Our loader is a dummy one which just spits out our source
    loader = ShowSourceLoader(modname, s)
    spec = spec_from_loader(modname, loader, origin = filename)
    module = module_from_spec(spec)
    # The code must be compiled so the function's code object has a filename
    code = compile(s, mode = 'exec', filename = filename)
    exec(code, module.__dict__)

    # `inspect.getmodule(...)`` requires it to be in sys.modules
    sys.modules[modname] = module

    return module.__dict__[fn_name]


from __future__ import annotations

import torch
from typing import Union, Sequence

from pyjuice.nodes import CircuitNodes


class Layer():

    propagation_alg_mapping = {
        "LL": 0,
        "MPE": 1,
        "GeneralLL": 2
    }

    def __init__(self, nodes: Sequence[CircuitNodes], disable_block_size_check: bool = False) -> None:

        # Nodes correspond to the current layer
        self.nodes = nodes

        # The set of unique scopes
        self.scopes = []
        for ns in self.nodes:
            if ns.scope not in self.scopes:
                self.scopes.append(ns.scope)

        if disable_block_size_check:
            self.block_size = None
        else:
            for i in range(1, len(nodes)):
                assert nodes[i].block_size == nodes[0].block_size, "`block_size` of nodes in the same layer must be identical."

            self.block_size = nodes[0].block_size

        self.device = torch.device("cpu")

    def enable_partial_evaluation(self, fw_scopes: Optional[Sequence[BitSet]] = None, bk_scopes: Optional[Sequence[BitSet]] = None):
        if not self.provided("fw_scope2localids") or not self.provided("bk_scope2localids"):
            raise ValueError("Please initialize node cache by calling `pc._create_scope2nid_cache()` first.")

        # Filter forward nodes
        if fw_scopes is not None:
            fw_partition_local_ids = [[] for _ in range(self.num_fw_partitions)]
            for scope in fw_scopes:
                if scope not in self.fw_scope2localids:
                    continue

                for partition_id, ids in enumerate(self.fw_scope2localids[scope]):
                    fw_partition_local_ids[partition_id].append(self.fw_scope2localids[scope][partition_id])

            self.fw_partition_local_ids = [
                torch.cat(ids, dim = 0) if len(ids) > 0 else torch.zeros([0], dtype = torch.long) for ids in fw_partition_local_ids
            ]

        # Filter backward nodes
        if bk_scopes is not None:
            bk_partition_local_ids = [[] for _ in range(self.num_bk_partitions)]
            for scope in bk_scopes:
                if scope not in self.bk_scope2localids:
                    continue

                for partition_id, ids in enumerate(self.bk_scope2localids[scope]):
                    bk_partition_local_ids[partition_id].append(self.bk_scope2localids[scope][partition_id])

            self.bk_partition_local_ids = [
                torch.cat(ids, dim = 0) if len(ids) > 0 else torch.zeros([0], dtype = torch.long) for ids in bk_partition_local_ids
            ]

    def disable_partial_evaluation(self, forward: bool = True, backward: bool = True):
        if forward:
            self.fw_partition_local_ids = None

        if backward:
            self.bk_partition_local_ids = None

    def provided(self, var_name):
        return hasattr(self, var_name) and getattr(self, var_name) is not None

    def is_sum(self):
        return False

    def is_prod(self):
        return False

    def is_input(self):
        return False

    def _get_propagation_alg_kwargs(self, propagation_alg: str, **kwargs):
        if propagation_alg == "LL":
            return {"alpha": 0.0}
        elif propagation_alg == "MPE":
            return {"alpha": 0.0}
        elif propagation_alg == "GeneralLL":
            return {"alpha": kwargs["alpha"]}
        else:
            raise ValueError(f"Unknown propagation algorithm {propagation_alg}.")



from __future__ import annotations

import torch
from copy import deepcopy as pydeepcopy
from typing import Optional, Dict

from pyjuice.nodes import CircuitNodes, InputNodes, ProdNodes, SumNodes
from pyjuice.utils import BitSet


def deepcopy(root_ns: CircuitNodes, tie_params: bool = False, max_block_size: Optional[int] = None,
             var_mapping: Optional[Dict[int,int]] = None) -> CircuitNodes:
    """
    Create a deepcopy of the input PC.

    :param root_ns: the input PC
    :type root_ns: CircuitNodes

    :param tie_params: whether to tie the parameters between the original PC and the copied PC (if tied, their parameters will always be the same)
    :type tie_params: bool

    :param max_block_size: the maximum block size of the copied PC
    :type max_block_size: Optional[int]

    :param var_mapping: a mapping dictionary between the variables of the original PC and the copied PC
    :type var_mapping: Optional[Dict[int,int]]

    :returns: a copied PC
    """

    assert not (max_block_size is not None and tie_params), "Could not change block size when `tie_params=True`."
    if max_block_size is not None:
        assert max_block_size > 0 and (max_block_size & (max_block_size - 1)) == 0, f"`max_block_size` must be a power of 2, but got `max_block_size={max_block_size}`."

    old2new = dict()
    tied_ns_pairs = []

    def dfs(ns: CircuitNodes):
        if ns in old2new:
            return

        # Recursively traverse children
        if ns.is_sum() or ns.is_prod():
            for cs in ns.chs:
                dfs(cs)

        new_chs = [old2new[cs] for cs in ns.chs]

        if not tie_params and ns.is_tied():
            tied_ns_pairs.append((ns, ns.get_source_ns()))

        if ns.is_sum():
            if not tie_params:
                if max_block_size is None:
                    edge_ids = ns.edge_ids.clone()
                    block_size = ns.block_size
                    params = ns.get_params()
                else:
                    old_ch_blk_size = ns.chs[0].block_size
                    old_blk_size = ns.block_size

                    new_ch_blk_size = new_chs[0].block_size
                    new_blk_size = min(old_blk_size, max_block_size)

                    blk_factor = old_blk_size // new_blk_size
                    ch_blk_factor = old_ch_blk_size // new_ch_blk_size

                    edge_ids = torch.stack(
                        (ns.edge_ids[0,:][:,None,None].repeat(1, blk_factor, ch_blk_factor) * blk_factor + torch.arange(0, blk_factor)[None,:,None],
                         ns.edge_ids[1,:][:,None,None].repeat(1, blk_factor, ch_blk_factor) * ch_blk_factor + torch.arange(0, ch_blk_factor)[None,None,:]),
                        dim = 0
                    ).flatten(1, 3)
                    block_size = new_blk_size

                    params = ns.get_params()
                    if params is not None:
                        num_edges = params.size(0)
                        params = params.reshape(num_edges, blk_factor, new_blk_size, ch_blk_factor, new_ch_blk_size).permute(0, 1, 3, 2, 4).flatten(0, 2)

                new_ns = SumNodes(
                    ns.num_nodes // block_size,
                    new_chs,
                    edge_ids,
                    block_size = block_size
                )
                if params is not None:
                    new_ns.set_params(params.clone(), normalize = False)
            else:
                new_ns = ns.duplicate(*new_chs, tie_params = True)
            
        elif ns.is_prod():
            if max_block_size is None:
                edge_ids = ns.edge_ids.clone()
                block_size = ns.block_size
            else:
                old_ch_blk_size = ns.chs[0].block_size
                old_blk_size = ns.block_size

                new_ch_blk_size = new_chs[0].block_size
                new_blk_size = min(old_blk_size, max_block_size)

                if old_blk_size == new_blk_size and old_ch_blk_size == new_ch_blk_size:
                    edge_ids = ns.edge_ids.clone()
                    block_size = ns.block_size
                else:
                    blk_factor = old_blk_size // new_blk_size
                    ch_blk_factor = old_ch_blk_size // new_ch_blk_size

                    if blk_factor == ch_blk_factor:
                        edge_ids = ns.edge_ids.clone()
                        edge_ids = edge_ids[:,None,:].repeat(1, blk_factor, 1) * blk_factor + torch.arange(0, blk_factor)[None,:,None]
                        edge_ids = edge_ids.flatten(0, 1)
                        block_size = new_blk_size
                    else:
                        raise NotImplementedError()

            new_ns = ProdNodes(
                ns.num_nodes // block_size,
                new_chs,
                edge_ids,
                block_size = block_size
            )
            
        else:
            assert ns.is_input()

            # Map variable scope
            if var_mapping is not None:
                ns_scope = ns.scope
                scope = BitSet()
                for v in ns_scope:
                    assert v in var_mapping, f"Variable {v} not found in `var_mapping`."
                    scope.add(var_mapping[v])
            else:
                scope = pydeepcopy(ns.scope)

            if max_block_size is None:
                block_size = ns.block_size
            else:
                block_size = min(ns.block_size, max_block_size)

            if not tie_params:
                new_ns = InputNodes(
                    num_node_blocks = ns.num_nodes // block_size,
                    scope = pydeepcopy(scope),
                    dist = pydeepcopy(ns.dist),
                    block_size = block_size
                )
                params = ns.get_params()
                if params is not None:
                    new_ns.set_params(params.clone(), normalize = False)
            else:
                new_ns = ns.duplicate(scope = scope, tie_params = True)

        old2new[ns] = new_ns

    dfs(root_ns)

    for ns, source_ns in tied_ns_pairs:
        new_ns = old2new[ns]
        new_source_ns = old2new[source_ns]

        new_ns._source_node = new_source_ns

    return old2new[root_ns]


import numpy as np
import torch

from typing import Union


def faiss_kmeans(data: Union[torch.Tensor,np.ndarray], n_clusters: int, niter: int = 100, nredo: int = 100):

    # Only import faiss if we run this function
    import faiss

    if isinstance(data, torch.Tensor):
        data = data.detach().cpu().numpy()
        convert_to_torch = True
    else:
        convert_to_torch = False

    kmeans = faiss.Clustering(data.shape[1], n_clusters)
    kmeans.verbose = False
    kmeans.niter = niter
    kmeans.nredo = nredo
    cfg = faiss.GpuIndexFlatConfig()
    cfg.useFloat16 = True
    cfg.device = self.device_id
    index = faiss.GpuIndexFlatL2(
        faiss.StandardGpuResources(),
        data.shape[1],
        cfg
    )
    kmeans.train(data, index)
    centroids = faiss.vector_float_to_array(kmeans.centroids).reshape(n_clusters, data.shape[1])

    index = faiss.IndexFlatL2(centroids.shape[1])
    index.add(np.ascontiguousarray(centroids))
    _, labels = index.search(data, 1)
    labels = labels.ravel()

    if convert_to_torch:
        labels = torch.from_numpy(labels)

    return labels


from __future__ import annotations

import numpy as np
import torch
import threading
import functools
import os
import warnings
import time
import triton
import triton.language as tl
from numba import njit
from copy import deepcopy
from typing import Optional, Sequence
from collections import OrderedDict

from pyjuice.nodes import CircuitNodes, SumNodes
from pyjuice.utils.kernel_launcher import FastJITFunction


## Helper functions ##


class OrderedSet():
    def __init__(self):
        self.item_set = set()
        self.item_list = list()

        self.index = 0

    def append(self, item):
        if item in self.item_set:
            return None
        
        self.item_set.add(item)
        self.item_list.append(item)

    def index(self, item):
        if item not in self.item_set:
            raise ValueError("Item not found.")

        return self.item_list.index(item)

    def __iter__(self):
        self.index = 0
        return self

    def __next__(self):
        if self.index < len(self.item_list):
            item = self.item_list[self.index]
            self.index += 1
            return item
        else:
            raise StopIteration  # To signal the end of iteration

    def __getitem__(self, idx):
        if idx >= len(self.item_list):
            raise ValueError()

        return self.item_list[idx]

    def __contains__(self, item):
        return item in self.item_set


def flatten_sum_nodes(ns: SumNodes, *args, use_cuda: bool = False):
    edge_ids = ns.edge_ids
    if use_cuda:
        edge_ids = edge_ids.cuda()

    if not ns.is_tied():
        return (ns.num_nodes, edge_ids, ns._param_range, [(c.num_nodes, c._output_ind_range) for c in ns.chs], *args)
    else:
        source_ns = ns.get_source_ns()
        return (ns.num_nodes, edge_ids, source_ns._param_range, [(c.num_nodes, c._output_ind_range) for c in ns.chs], *args)


def get_chunk_ids(n, k):
    chunk_size = n // k
    remainder = n % k

    chunks = []
    start = 0
    for i in range(k):
        if i < remainder:
            end = start + chunk_size + 1
        else:
            end = start + chunk_size

        chunks.append((start, end))
        start = end

    return chunks


def next_power_of_2(x: torch.Tensor):
    return torch.where(
        x == 1,
        1,
        2 ** torch.ceil(torch.log2(x.float())).long()
    )


## Compilation for SumLayer ##


def get_sum_layer_forward_stats(nodes: Sequence[SumNodes], global_nid_start: int):
    layer_num_nblocks = sum(map(lambda ns: ns.num_node_blocks, nodes))
    layer_num_edges = 0

    n_sid = 0
    n_chs = torch.zeros([layer_num_nblocks], dtype = torch.long)
    for ns_idx, ns in enumerate(nodes):
        n_eid = n_sid + ns.num_node_blocks

        curr_n_chs = torch.bincount(ns.edge_ids[0,:])
        # To maximize flexibility, we point to individual child nodes instead of a node block
        n_chs[n_sid:n_eid] = curr_n_chs * ns.ch_block_size

        ns._output_ind_range = (global_nid_start, global_nid_start + ns.num_nodes)
        global_nid_start += ns.num_nodes
        layer_num_edges += ns.num_edges

        n_sid = n_eid

    return layer_num_nblocks, layer_num_edges, n_chs


def get_sum_layer_backward_stats(nodes: Sequence[SumNodes]):
    ch_gsize2cs = dict()
    ch_gsize2num_nblocks = dict()
    cs2parns = dict()

    for ns in nodes:
        for cs in ns.chs:
            ch_gsize = cs.block_size

            if ch_gsize not in ch_gsize2cs:
                ch_gsize2cs[ch_gsize] = OrderedSet()
                ch_gsize2num_nblocks[ch_gsize] = 0

            if cs not in ch_gsize2cs[ch_gsize]:
                ch_gsize2cs[ch_gsize].append(cs)
                ch_gsize2num_nblocks[ch_gsize] += cs.num_node_blocks

            if cs not in cs2parns:
                cs2parns[cs] = OrderedSet()

            cs2parns[cs].append(ns)

    # Iterate over all child nodes to get the parent (# node blocks) counts
    ch_gsize2n_pargs = dict()
    for ch_gsize, ch_nodes in ch_gsize2cs.items():
        n_sid = 0
        n_pargs = torch.zeros([ch_gsize2num_nblocks[ch_gsize]], dtype = torch.long)
        for cs in ch_nodes:
            n_eid = n_sid + cs.num_node_blocks

            pargcounts = torch.zeros([cs.num_node_blocks], dtype = torch.long)
            for ns in cs2parns[cs]:
                cs_id = ns.chs.index(cs)
                edge_sid = sum([c.num_node_blocks for c in ns.chs[:cs_id]])
                edge_eid = edge_sid + cs.num_node_blocks

                criterion = (ns.edge_ids[1,:] >= edge_sid) & (ns.edge_ids[1,:] < edge_eid)
                pargcounts += torch.bincount(ns.edge_ids[1,criterion] - edge_sid, minlength = cs.num_node_blocks)

            n_pargs[n_sid:n_eid] = pargcounts

            n_sid = n_eid

        ch_gsize2n_pargs[ch_gsize] = n_pargs

    return ch_gsize2cs, ch_gsize2num_nblocks, ch_gsize2n_pargs, cs2parns


@njit
def _assign_chid_kernel(chs_offsets, ns_nchs, edge_ids):
    for i in range(edge_ids.shape[1]):
        nid = edge_ids[0,i]
        idx = ns_nchs[nid]
        chs_offsets[i] = idx
        ns_nchs[nid] = idx + 1


@triton.jit
def _assign_target_ncpids_kernel(target_nids_ptr, nids_partition_start_ptr, target_cids_ptr, pcids_partition_start_ptr,
                                 target_pids_ptr, target_pfids_ptr, edge_ids_ptr, chs_offsets_ptr, n_partition_ids_ptr, 
                                 n_id_in_partition_ptr, cs_ele_id_start_ptr, cs_node_cum_ids_ptr, fw_partition_max_chs_ptr, 
                                 cum_n_chs_ptr, ns_param_ids_ptr, ns_param_flow_ids_ptr, cid_node_id_ptr, constexprs_ptr, 
                                 num_chs: tl.constexpr, num_chs_np2: tl.constexpr, add_params_flag: tl.constexpr, 
                                 add_param_flows_flag: tl.constexpr, BLOCK_SIZE: tl.constexpr):

    pid = tl.program_id(axis = 0)
    block_start = pid * BLOCK_SIZE

    # Retrieve all constexprs
    global_nid_start = tl.load(constexprs_ptr)
    ns_pid_start = tl.load(constexprs_ptr + 1)
    ns_pfid_start = tl.load(constexprs_ptr + 2)
    nblock_start = tl.load(constexprs_ptr + 3)
    num_edges = tl.load(constexprs_ptr + 4)
    block_size = tl.load(constexprs_ptr + 5)

    # Get edge indices to be processed by the current block
    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < num_edges

    # Get `nid` and `cid` (size of `edge_ids` is [2, num_edges])
    nid = tl.load(edge_ids_ptr + offsets, mask = mask, other = 0)
    cid = tl.load(edge_ids_ptr + offsets + num_edges, mask = mask, other = 0)

    # Get `partition_id` and `local_id`
    partition_id = tl.load(n_partition_ids_ptr + nid + nblock_start, mask = mask, other = 0)
    local_id = tl.load(n_id_in_partition_ptr + nid + nblock_start, mask = mask, other = 0)

    # Get the child ns index every `cid` belongs to and the cum nodes & global sid
    cs_offsets = tl.arange(0, num_chs_np2)
    cs_node_cum_ids = tl.load(cs_node_cum_ids_ptr + cs_offsets, mask = (cs_offsets < num_chs), other = 0)
    
    # Get the `cs` indices the edges belong to
    cid_node_id = tl.load(cid_node_id_ptr + offsets, mask = mask, other = 0)

    cs_cum_num = tl.load(cs_node_cum_ids_ptr + cid_node_id, mask = mask, other = 0)
    cs_ele_ind = tl.load(cs_ele_id_start_ptr + cid_node_id, mask = mask, other = 0)

    # Get child offsets
    # Note: this is the `?` mark in `cids[partition_id][local_id,?]`
    chs_offset = tl.load(chs_offsets_ptr + offsets, mask = mask, other = 0)

    # Store to `target_nids`
    nids_start = tl.load(nids_partition_start_ptr + partition_id, mask = mask, other = 0)
    global_nid = global_nid_start + (nblock_start + nid) * block_size
    tl.store(target_nids_ptr + nids_start + local_id, global_nid, mask = mask)

    # Store to `target_cids`
    partition_max_n_chs = tl.load(fw_partition_max_chs_ptr + partition_id, mask = mask, other = 0)
    pcids_start = tl.load(pcids_partition_start_ptr + partition_id, mask = mask, other = 0)
    pcids_offsets = pcids_start + local_id * partition_max_n_chs + chs_offset
    global_cid = cs_ele_ind + cid - cs_cum_num
    tl.store(target_cids_ptr + pcids_offsets, global_cid, mask = mask)

    # Store to `target_pids`
    ns_local_pid = tl.load(cum_n_chs_ptr + nid, mask = mask, other = 0)
    global_pid = ns_pid_start + (ns_local_pid + chs_offset) * block_size
    tl.store(target_pids_ptr + pcids_offsets, global_pid, mask = mask)

    # Store to `target_pfids`
    global_pfid = ns_pfid_start + (ns_local_pid + chs_offset) * block_size
    tl.store(target_pfids_ptr + pcids_offsets, global_pfid, mask = mask)

    # Global parameter indices for all edges
    if add_params_flag:
        tl.store(ns_param_ids_ptr + offsets, global_pid, mask = mask)

    # Global parameter flow indices for all edges
    if add_param_flows_flag:
        tl.store(ns_param_flow_ids_ptr + offsets, global_pfid, mask = mask)


@torch.no_grad()
def sum_layer_forward_compilation(nodes, fw_partition_max_chs, n_partition_ids, n_id_in_partition, num_ngs_in_partition, n_chs,
                                  global_nid_start: int, global_pid_start: int, global_pfid_start: int, node2tiednodes: dict, 
                                  max_tied_ns_per_parflow_block: int = 4, use_cuda: bool = True):

    if use_cuda and not torch.cuda.is_available():
        use_cuda = False

    if use_cuda:
        device = torch.device(f"cuda:{torch.cuda.current_device()}")
    else:
        device = torch.device("cpu")

    if use_cuda:
        # We construct a flattened version of `nids` where the vectors of every partition is concatenated
        # into a single vector. `nids_block_start` is used to indicate the start index of every block's
        # `nids`. That is, `target_nids[nids_partition_start[i]:nids_partition_start[i+1]] == nids[i]`
        nids_partition_start = torch.zeros_like(num_ngs_in_partition)
        nids_partition_start[1:] = torch.cumsum(num_ngs_in_partition[:-1], dim = 0)
        target_nids = torch.zeros([num_ngs_in_partition.sum()], dtype = torch.long).to(device)

        # Similarly, we flatten `cids`...
        # Note: we call it `pcids...` because it is shared with `target_pids`
        pcids_partition_start = torch.zeros_like(num_ngs_in_partition)
        pcids_partition_start[1:] = torch.cumsum((num_ngs_in_partition * fw_partition_max_chs)[:-1], dim = 0)
        target_cids = torch.zeros([(num_ngs_in_partition * fw_partition_max_chs).sum()], dtype = torch.long).to(device)

        # ...and `pids`
        target_pids = torch.zeros([(num_ngs_in_partition * fw_partition_max_chs).sum()], dtype = torch.long).to(device)

        # ...and `pfids`
        target_pfids = torch.zeros([(num_ngs_in_partition * fw_partition_max_chs).sum()], dtype = torch.long).to(device)

        # Move necessary tensors to GPU
        n_partition_ids = n_partition_ids.to(device)
        n_id_in_partition = n_id_in_partition.to(device)
        fw_partition_max_chs = fw_partition_max_chs.to(device)

    else:
        nids = [torch.zeros([num_ngs_in_partition[i]], dtype = torch.long) for i in range(len(num_ngs_in_partition))]
        cids = [torch.zeros([num_ngs_in_partition[i], fw_partition_max_chs[i]], dtype = torch.long) for i in range(len(num_ngs_in_partition))]
        pids = [torch.zeros([num_ngs_in_partition[i], fw_partition_max_chs[i]], dtype = torch.long) for i in range(len(num_ngs_in_partition))]
        pfids = [torch.zeros([num_ngs_in_partition[i], fw_partition_max_chs[i]], dtype = torch.long) for i in range(len(num_ngs_in_partition))]

        ngid_in_partition = torch.zeros([len(num_ngs_in_partition)], dtype = torch.long)

    all_ns_param_ids = dict()
    all_ns_param_flow_ids = dict()
    original_param_nids = [] # `ns` with their original parameters (i.e., not tied)

    # This is the main loop: iterate over `ns` in the layer
    nblock_start = 0 # The start index of the node blocks in the current `ns`
    node2tiedcounts = dict() # Locally accumulate the occupation count
    for ns_idx, ns in enumerate(nodes):

        if not ns.is_tied():
            if not ns.provided("_param_range"):
                global_pid_end = global_pid_start + ns.num_edges
                ns._param_range = (global_pid_start, global_pid_end)
                global_pid_start = global_pid_end

                global_pfid_end = global_pfid_start + ns.num_edges
                ns._param_flow_range = (global_pfid_start, global_pfid_end)
                global_pfid_start = global_pfid_end

                add_params_flag = True
                add_param_flows_flag = True
            else:
                assert ns.provided("_param_flow_range")

                add_params_flag = False
                add_param_flows_flag = False

            original_param_nids.append(ns_idx)
                
            # Global pid and pfid start index for `ns`
            ns_pid_start = ns._param_range[0]
            ns_pfid_start = ns._param_flow_range[0]
        else:
            source_ns = ns.get_source_ns()

            # Initialize parameters
            if not source_ns.provided("_param_range"):
                global_pid_end = global_pid_start + ns.num_edges
                ns._param_range = (global_pid_start, global_pid_end)

                global_pfid_end = global_pfid_start + ns.num_edges
                ns._param_flow_range = (global_pfid_start, global_pfid_end)

                source_ns._param_range = (global_pid_start, global_pid_end)
                source_ns._param_flow_range = (global_pfid_start, global_pfid_end)

                global_pid_start = global_pid_end
                global_pfid_start = global_pfid_end

                add_params_flag = True
                add_param_flows_flag = True
            else:
                ns._param_range = deepcopy(source_ns._param_range)

                add_params_flag = False
                add_param_flows_flag = False

            if source_ns not in node2tiednodes:
                node2tiednodes[source_ns] = [[source_ns], [source_ns._param_flow_range]]
                node2tiedcounts[source_ns] = [1]
            elif source_ns not in node2tiedcounts:
                node2tiedcounts[source_ns] = [0 for _ in range(len(node2tiednodes[source_ns][0]))]
            
            if not ns.provided("_param_flow_range"):
                if all([dup_count >= max_tied_ns_per_parflow_block for dup_count in node2tiedcounts[source_ns]]):
                    global_pfid_end = global_pfid_start + ns.num_edges
                    ns._param_flow_range = (global_pfid_start, global_pfid_end)
                    global_pfid_start = global_pfid_end
                    node2tiednodes[source_ns][1].append(ns._param_flow_range)

                    node2tiednodes[source_ns][0].append(ns)
                    node2tiedcounts[source_ns].append(1)

                    add_param_flows_flag = True
                else:
                    target_id = min(range(len(node2tiedcounts[source_ns])), key = lambda i: node2tiedcounts[source_ns][i])
                    ns._param_flow_range = deepcopy(node2tiednodes[source_ns][1][target_id])

                    node2tiedcounts[source_ns][target_id] += 1

                    add_param_flows_flag = False

            # Global pid and pfid start index for `ns`
            ns_pid_start = source_ns._param_range[0]
            ns_pfid_start = ns._param_flow_range[0]

        # number of node blocks
        ns_num_nblocks = ns.num_node_blocks

        # Edge indices of size [2, ns_num_edges]
        # Here child ids of the edges are flattened out, i.e., every edge points to 
        # an actual "node" instead of a node block
        edge_ids = ns.edge_ids.clone()
        edge_ids = edge_ids[:,:,None].repeat(1, 1, ns.ch_block_size)
        edge_ids[1,:,:] *= ns.ch_block_size
        edge_ids[1,:,:] += torch.arange(0, ns.ch_block_size)[None,:]
        edge_ids = edge_ids.reshape(2, ns.edge_ids.size(1) * ns.ch_block_size).contiguous()
        ns_num_edges = edge_ids.size(1)

        if use_cuda:
            ## GPU mode ##

            # Precompute the child offset ids for every edge. That is, the `?` 
            # mark in `cids[partition][local_id,?]`
            chs_offsets = np.zeros([ns_num_edges], dtype = np.int64)
            ns_nchs = np.zeros([ns_num_nblocks], dtype = np.int64)

            _assign_chid_kernel(chs_offsets, ns_nchs, edge_ids.numpy())
            chs_offsets = torch.from_numpy(chs_offsets)

            # Construct helper indices for child nodes
            # `cs_ele_id_start` contains the global start indices of the child nodes
            # `cs_node_cum_ids` contains the local cumulative number of child nodes
            cs_ele_id_start = torch.zeros([ns.num_chs], dtype = torch.long)
            cs_node_cum_ids = torch.zeros([ns.num_chs], dtype = torch.long)
            for i, cs in enumerate(ns.chs):
                cs_ele_id_start[i] = cs._output_ind_range[0]
                if i < ns.num_chs - 1:
                    cs_node_cum_ids[i+1] = cs_node_cum_ids[i] + cs.num_nodes

            # Cumulative nchs
            ns_nchs = torch.from_numpy(ns_nchs)
            cum_n_chs = torch.zeros([ns_num_nblocks], dtype = torch.long)
            cum_n_chs[1:] = torch.cumsum(ns_nchs[:-1], dim = 0)

            if add_params_flag:
                ns_param_ids = torch.zeros([ns_num_edges], dtype = torch.long).to(device)
            else:
                ns_param_ids = None

            if add_param_flows_flag:
                ns_param_flow_ids = torch.zeros([ns_num_edges], dtype = torch.long).to(device)
            else:
                ns_param_flow_ids = None

            # The following kernel assigns the corresponding indices to `nids`, `cids`, and `pids`
            # We first move necessary buffers to GPU
            nids_partition_start = nids_partition_start.to(device)
            edge_ids = edge_ids.to(device)
            chs_offsets = chs_offsets.to(device)
            cs_ele_id_start = cs_ele_id_start.to(device)
            cs_node_cum_ids = cs_node_cum_ids.to(device)
            cum_n_chs = cum_n_chs.to(device)
            pcids_partition_start = pcids_partition_start.to(device)

            # Which `cs` are the edges pointing to
            cid_node_id = (edge_ids[1,:].unsqueeze(1) >= cs_node_cum_ids[None,:]).sum(dim = 1) - 1

            # We store these constants in a tensor and retrieve them in the kernel
            # This is to avoid `triton` from compiling separate kernels for every layer configuration
            # Saves 99.9% compilation time :)
            constexprs = torch.tensor([global_nid_start, ns_pid_start, ns_pfid_start, nblock_start, ns_num_edges, ns.block_size]).long().to(device)

            num_chs_np2 = triton.next_power_of_2(ns.num_chs)

            # Make the grid and launch kernel
            grid = lambda meta: (triton.cdiv(ns_num_edges, meta["BLOCK_SIZE"]),)

            _assign_target_ncpids_kernel[grid](
                target_nids, nids_partition_start, target_cids, pcids_partition_start,
                target_pids, target_pfids, edge_ids, chs_offsets, n_partition_ids, 
                n_id_in_partition, cs_ele_id_start, cs_node_cum_ids, fw_partition_max_chs, 
                cum_n_chs, ns_param_ids, ns_param_flow_ids, cid_node_id, constexprs, ns.num_chs, num_chs_np2, 
                add_params_flag, add_param_flows_flag, BLOCK_SIZE = min(2048, 2**20 // num_chs_np2)
            )

            nblock_start += ns_num_nblocks

        else:
            ## CPU mode ##

            # Get number of child nodes for all nodes
            ns_nchs = torch.bincount(edge_ids[0,:], minlength = ns_num_nblocks)

            cs_node_cum_nodes = torch.zeros([ns.num_chs], dtype = torch.long)
            cs_node_cum_nodes[0] = ns.chs[0].num_nodes
            for i in range(1, ns.num_chs):
                cs_node_cum_nodes[i] = cs_node_cum_nodes[i-1] + ns.chs[i].num_nodes

            if add_params_flag:
                ns_param_ids = torch.zeros([ns_num_edges], dtype = torch.long)
            else:
                ns_param_ids = None

            if add_param_flows_flag:
                ns_param_flow_ids = torch.zeros([ns_num_edges], dtype = torch.long).to(device)
            else:
                ns_param_flow_ids = None

            # Iterate over node blocks
            cum_n_chs = 0
            for ng_id in range(ns_num_nblocks):
                partition_id = (ns_nchs[ng_id] > fw_partition_max_chs).sum()
                local_id = ngid_in_partition[partition_id]

                global_nid = ns._output_ind_range[0] + ng_id * ns.block_size

                # Assign `nids`
                nids[partition_id][local_id] = global_nid

                # Assign `cids`
                criterion = (edge_ids[0,:] == ng_id)
                local_cids = edge_ids[1,criterion]
                cids_gid = (local_cids[:,None] >= cs_node_cum_nodes[None,:]).sum(dim = 1)
                for ch_id in range(local_cids.size(0)):
                    local_base = cs_node_cum_nodes[cids_gid[ch_id]-1] if cids_gid[ch_id] >= 1 else 0
                    global_cid = ns.chs[cids_gid[ch_id]]._output_ind_range[0] + local_cids[ch_id] - local_base
                    cids[partition_id][local_id, ch_id] = global_cid

                # Assign `pids`
                global_pids = ns_pid_start + cum_n_chs + torch.arange(0, ns.block_size * criterion.sum(), ns.block_size)
                pids[partition_id][local_id, 0:global_pids.size(0)] = global_pids

                # Assign `pfids`
                global_pfids = ns_pfid_start + cum_n_chs + torch.arange(0, ns.block_size * criterion.sum(), ns.block_size)
                pfids[partition_id][local_id, 0:global_pfids.size(0)] = global_pfids

                cum_n_chs += ns.block_size * criterion.sum()

                if add_params_flag:
                    ns_param_ids[criterion] = global_pids

                if add_param_flows_flag:
                    ns_param_flow_ids[criterion] = global_pfids

                ngid_in_partition[partition_id] = local_id + 1

        if add_params_flag:
            all_ns_param_ids[ns_idx] = ns_param_ids

        if add_param_flows_flag:
            all_ns_param_flow_ids[ns_idx] = ns_param_flow_ids

    # Store global -> local parameter id mapping in `ns`
    for ns_idx, param_ids in all_ns_param_ids.items():
        ns = nodes[ns_idx]
        if ns.is_tied():
            ns = ns.get_source_ns()
        # Every edge specify the start id of [ch_block_size, block_size] parameters
        ns._param_ids = param_ids.cpu()[0::ns.ch_block_size]

    # Store global -> local parameter flow id mapping in `ns`
    for ns_idx, param_flow_ids in all_ns_param_flow_ids.items():
        ns = nodes[ns_idx]
        # Every edge specify the start id of [ch_block_size, block_size] parameter flows
        ns._param_flow_ids = param_flow_ids.cpu()[0::ns.ch_block_size]

    # Store local -> global parameter id mapping in `ns`
    for ns_idx in original_param_nids:
        ns = nodes[ns_idx]
        ns._inverse_param_ids = torch.argsort(ns._param_ids)

    if use_cuda:
        # Restore `nids`
        target_nids = target_nids.cpu()
        nids = []
        for partition_id in range(num_ngs_in_partition.size(0)):
            sid = nids_partition_start[partition_id]
            eid = sid + num_ngs_in_partition[partition_id]
            nids.append(target_nids[sid:eid].contiguous())

        # Restore `cids` and `pids`
        target_cids = target_cids.cpu()
        target_pids = target_pids.cpu()
        target_pfids = target_pfids.cpu()
        cids = []
        pids = []
        pfids = []
        for partition_id in range(num_ngs_in_partition.size(0)):
            sid = pcids_partition_start[partition_id]
            gsize = num_ngs_in_partition[partition_id]
            gnchs = fw_partition_max_chs[partition_id]
            eid = sid + gsize * gnchs
            cids.append(target_cids[sid:eid].reshape(gsize, gnchs).contiguous())
            pids.append(target_pids[sid:eid].reshape(gsize, gnchs).contiguous())
            pfids.append(target_pfids[sid:eid].reshape(gsize, gnchs).contiguous())

    return nids, cids, pids, pfids, global_pid_start, global_pfid_start


@njit
def _assign_chid_kernel(chs_offsets, ns_nchs, edge_ids):
    for i in range(edge_ids.shape[1]):
        nid = edge_ids[0,i]
        idx = ns_nchs[nid]
        chs_offsets[i] = idx
        ns_nchs[nid] = idx + 1


@njit
def _assign_parid_kernel(pars_offsets, cs_npars, edge_ids, edge_sid):
    for i in range(edge_ids.shape[1]):
        cid = edge_ids[1,i]
        idx = cs_npars[cid]
        pars_offsets[edge_sid+i] = idx
        cs_npars[cid] = idx + 1


@triton.jit
def _assign_target_chpapids_kernel(target_chids_ptr, chids_partition_start_ptr, target_parids_ptr, target_parpids_ptr, 
                                   parids_partition_start_ptr, edge_ids_ptr, pars_offsets_ptr, n_partition_ids_ptr, 
                                   n_id_in_partition_ptr, num_ngs_in_partition_ptr, partition_max_pars_ptr, cum_n_chs_ptr, 
                                   chs_offsets_ptr, constexprs_ptr, BLOCK_SIZE: tl.constexpr):
    
    pid = tl.program_id(axis = 0)
    block_start = pid * BLOCK_SIZE

    # Retrieve all constexprs
    ns_global_node_start = tl.load(constexprs_ptr)
    cs_global_ele_start = tl.load(constexprs_ptr + 1)
    ns_block_size = tl.load(constexprs_ptr + 2)
    cs_block_size = tl.load(constexprs_ptr + 3)
    ns_pid_start = tl.load(constexprs_ptr + 4)
    num_edges = tl.load(constexprs_ptr + 5)
    cs_nblock_start = tl.load(constexprs_ptr + 6)
    pars_offsets_start = tl.load(constexprs_ptr + 7)

    # Get edge indices to be processed by the current block
    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < num_edges

    # Get `cid` and `nid` (size of `edge_ids` is [2, num_edges])
    cid = tl.load(edge_ids_ptr + offsets + num_edges, mask = mask, other = 0)
    nid = tl.load(edge_ids_ptr + offsets, mask = mask, other = 0)

    # Get `partition_id` and `local_id`
    partition_id = tl.load(n_partition_ids_ptr + cid + cs_nblock_start, mask = mask, other = 0)
    local_id = tl.load(n_id_in_partition_ptr + cid + cs_nblock_start, mask = mask, other = 0)

    # Get parent offsets
    # Note: this is the `?` mark in `parids[partition_id][local_id,?]`
    pars_offset = tl.load(pars_offsets_ptr + pars_offsets_start + offsets, mask = mask, other = 0)

    # Store to `target_chids`
    chids_start = tl.load(chids_partition_start_ptr + partition_id, mask = mask, other = 0)
    global_chid = cs_global_ele_start + cid * cs_block_size
    tl.store(target_chids_ptr + chids_start + local_id, global_chid, mask = mask)

    # Store to `target_parids`
    partition_max_n_pargs = tl.load(partition_max_pars_ptr + partition_id, mask = mask, other = 0)
    parids_start = tl.load(parids_partition_start_ptr + partition_id, mask = mask, other = 0)
    parids_offsets = parids_start + local_id * partition_max_n_pargs + pars_offset
    global_parid = ns_global_node_start + nid * ns_block_size
    tl.store(target_parids_ptr + parids_offsets, global_parid, mask = mask)

    # Store to `target_parpids`
    ns_local_pid = tl.load(cum_n_chs_ptr + nid, mask = mask, other = 0)
    chs_offset = tl.load(chs_offsets_ptr + offsets, mask = mask, other = 0)
    global_pid = ns_pid_start + (ns_local_pid + chs_offset) * ns_block_size * cs_block_size
    tl.store(target_parpids_ptr + parids_offsets, global_pid, mask = mask)


def _assign_target_chpapids_cpu(target_chids, chids_partition_start, target_parids, target_parpids, parids_partition_start,
                                edge_ids, pars_offsets, n_partition_ids, n_id_in_partition, num_ngs_in_partition,
                                partition_max_pars, cum_n_chs, chs_offsets, ns_global_node_start, cs_global_ele_start, 
                                ns_block_size, cs_block_size, ns_pid_start, ns_num_edges, cs_nblock_start, pars_offsets_start):

    for edge_id in range(ns_num_edges):
        # Get `cid` and `nid` (size of `edge_ids` is [2, num_edges])
        cid = edge_ids[1, edge_id]
        nid = edge_ids[0, edge_id]

        # Get `partition_id` and `local_id`
        partition_id = n_partition_ids[cid + cs_nblock_start]
        local_id = n_id_in_partition[cid + cs_nblock_start]

        # Get parent offsets
        # Note: this is the `?` mark in `parids[partition_id][local_id,?]`
        pars_offset = pars_offsets[pars_offsets_start + edge_id]

        # Store to `target_chids`
        chids_start = chids_partition_start[partition_id]
        global_chid = cs_global_ele_start + cid * cs_block_size
        target_chids[chids_start + local_id] = global_chid

        # Store to `target_parids`
        partition_max_n_pargs = partition_max_pars[partition_id]
        parids_start = parids_partition_start[partition_id]
        parids_offsets = parids_start + local_id * partition_max_n_pargs + pars_offset
        global_parid = ns_global_node_start + nid * ns_block_size
        target_parids[parids_offsets] = global_parid

        # Store to `target_parpids`
        ns_local_pid = cum_n_chs[nid]
        chs_offset = chs_offsets[edge_id]
        global_pid = ns_pid_start + (ns_local_pid + chs_offset) * ns_block_size * cs_block_size
        target_parpids[parids_offsets] = global_pid


@torch.no_grad()
def sum_layer_backward_compilation(nodes, cs2parns, n_partition_ids, n_id_in_partition, num_ngs_in_partition, partition_max_pars,
                                   use_cuda: bool = False):

    if use_cuda and not torch.cuda.is_available():
        use_cuda = False

    if use_cuda:
        device = torch.device(f"cuda:{torch.cuda.current_device()}")
    else:
        device = torch.device("cpu")

    # We construct a flattened version of `chids` where the vectors of every partition is concatenated
    # into a single vector. `chids_partition_start` is used to indicate the start index of every partition's
    # `chids`. That is, `target_chids[chids_partition_start[i]:chids_partition_start[i+1]] == chids[i]`
    chids_partition_start = torch.zeros_like(num_ngs_in_partition)
    chids_partition_start[1:] = torch.cumsum(num_ngs_in_partition[:-1], dim = 0)
    target_chids = torch.zeros([num_ngs_in_partition.sum()], dtype = torch.long).to(device)

    # Similarly, we flatten `parids`...
    # Note: we call it `pcids...` because it is shared with `target_pids`
    parids_partition_start = torch.zeros_like(num_ngs_in_partition)
    parids_partition_start[1:] = torch.cumsum((num_ngs_in_partition * partition_max_pars)[:-1], dim = 0)
    target_parids = torch.zeros([(num_ngs_in_partition * partition_max_pars).sum()], dtype = torch.long).to(device)

    # ...and `parpids`
    target_parpids = torch.zeros([(num_ngs_in_partition * partition_max_pars).sum()], dtype = torch.long).to(device)

    # Move tensors to GPU
    n_partition_ids = n_partition_ids.to(device)
    n_id_in_partition = n_id_in_partition.to(device)
    num_ngs_in_partition = num_ngs_in_partition.to(device)
    partition_max_pars = partition_max_pars.to(device)

    # This is the main loop: iterate over `cs` in the layer
    cs_nblock_start = 0 # The start index of nodes in the current `cs`
    ns2cum_n_chs = dict()
    ns2chs_offsets = dict()
    for cs in nodes:

        # Collect all edge ids that point to `cs` in every parent `ns`
        par_edge_ids = []
        local_ns2chs_offsets = dict()
        for ns in cs2parns[cs]:
            cs_id = ns.chs.index(cs)
            edge_sid = sum([c.num_node_blocks for c in ns.chs[:cs_id]])
            edge_eid = edge_sid + cs.num_node_blocks

            criterion = (ns.edge_ids[1,:] >= edge_sid) & (ns.edge_ids[1,:] < edge_eid)
            extracted_edge_ids = ns.edge_ids[:,criterion].clone()
            extracted_edge_ids[1,:] -= edge_sid

            par_edge_ids.append(extracted_edge_ids)

            # Recreate `chs_offsets` and `cum_n_chs` to get compute the parameter ids
            if not ns in ns2cum_n_chs:
                chs_offsets = np.zeros([ns.edge_ids.size(1)], dtype = np.int64)
                ns_nchs = np.zeros([ns.num_node_blocks], dtype = np.int64)

                _assign_chid_kernel(chs_offsets, ns_nchs, ns.edge_ids.numpy())
                chs_offsets = torch.from_numpy(chs_offsets)

                ns_nchs = torch.from_numpy(ns_nchs)
                cum_n_chs = torch.zeros([ns.num_node_blocks], dtype = torch.long)
                cum_n_chs[1:] = torch.cumsum(ns_nchs[:-1], dim = 0)

                ns2cum_n_chs[ns] = cum_n_chs
                ns2chs_offsets[ns] = chs_offsets

            local_ns2chs_offsets[ns] = ns2chs_offsets[ns][criterion]

        cs_num_nblocks = cs.num_node_blocks
        cs_num_edges = sum([edge_ids.size(1) for edge_ids in par_edge_ids])
        
        # Precompute the parent offset ids for every. That is, the `?`
        # mark in `parids[partition_id][local_id,?]`
        pars_offsets = np.zeros([cs_num_edges], dtype = np.int64)
        cs_npars = np.zeros([cs_num_nblocks], dtype = np.int64)

        edge_sid = 0
        for edge_ids in par_edge_ids:
            edge_eid = edge_sid + edge_ids.size(1)

            _assign_parid_kernel(pars_offsets, cs_npars, edge_ids.numpy(), edge_sid)

            edge_sid = edge_eid

        pars_offsets = torch.from_numpy(pars_offsets)

        # Move necessary buffers to GPU
        chids_partition_start = chids_partition_start.to(device)
        parids_partition_start = parids_partition_start.to(device)
        pars_offsets = pars_offsets.to(device)

        pars_offsets_start = 0
        for ns, edge_ids in zip(cs2parns[cs], par_edge_ids):

            ns_num_edges = edge_ids.size(1)
            edge_ids = edge_ids.to(device)

            if ns.is_tied():
                ns_pid_start = ns.get_source_ns()._param_range[0]
            else:
                ns_pid_start = ns._param_range[0]

            # Get `cum_n_chs` and `chs_offsets`, which are used to get the parameter indices
            cum_n_chs = ns2cum_n_chs[ns].to(device)
            chs_offsets = local_ns2chs_offsets[ns].to(device)

            # We store these constants in a tensor and retrieve them in the kernel
            # This is to avoid `triton` from compiling separate kernels for every layer configuration
            # Saves 99.9% compilation time :)
            cs_global_ele_start = cs._output_ind_range[0]
            ns_global_node_start = ns._output_ind_range[0]
            ns_block_size = ns.block_size
            cs_block_size = cs.block_size
            
            if use_cuda:
                constexprs = torch.tensor([ns_global_node_start, cs_global_ele_start, ns_block_size, cs_block_size, 
                                           ns_pid_start, ns_num_edges, cs_nblock_start, pars_offsets_start]).long().to(device)

                # Make the grid and launch kernel
                grid = lambda meta: (triton.cdiv(ns_num_edges, meta["BLOCK_SIZE"]),)

                _assign_target_chpapids_kernel[grid](
                    target_chids, chids_partition_start, target_parids, target_parpids, parids_partition_start,
                    edge_ids, pars_offsets, n_partition_ids, n_id_in_partition, num_ngs_in_partition,
                    partition_max_pars, cum_n_chs, chs_offsets, constexprs, BLOCK_SIZE = 1024
                )
            else:
                _assign_target_chpapids_cpu(
                    target_chids, chids_partition_start, target_parids, target_parpids, parids_partition_start,
                    edge_ids, pars_offsets, n_partition_ids, n_id_in_partition, num_ngs_in_partition,
                    partition_max_pars, cum_n_chs, chs_offsets, ns_global_node_start, cs_global_ele_start, 
                    ns_block_size, cs_block_size, ns_pid_start, ns_num_edges, cs_nblock_start, pars_offsets_start
                )

            pars_offsets_start += ns_num_edges

        cs_nblock_start += cs.num_node_blocks

    # Restore `chids`
    target_chids = target_chids.cpu()
    chids = []
    for partition_id in range(num_ngs_in_partition.size(0)):
        sid = chids_partition_start[partition_id]
        eid = sid + num_ngs_in_partition[partition_id]
        chids.append(target_chids[sid:eid].contiguous())

    # Restore `parids` and `parpids`
    target_parids = target_parids.cpu()
    target_parpids = target_parpids.cpu()
    parids = []
    parpids = []
    for partition_id in range(num_ngs_in_partition.size(0)):
        sid = parids_partition_start[partition_id]
        gsize = num_ngs_in_partition[partition_id]
        gnchs = partition_max_pars[partition_id]
        eid = sid + gsize * gnchs
        parids.append(target_parids[sid:eid].reshape(gsize, gnchs).contiguous())
        parpids.append(target_parpids[sid:eid].reshape(gsize, gnchs).contiguous())

    return chids, parids, parpids


## Compilation for ProdLayer ##

def get_prod_layer_stats(nodes: Sequence[SumNodes], block_size: int, global_nid_start: int, use_block_sparse_edges: bool):
    if use_block_sparse_edges:
        layer_num_nblock = sum(map(lambda ns: ns.num_node_blocks, nodes))
        layer_num_edges = 0

        ng_sid = 0
        n_chgs = torch.zeros([layer_num_nblock], dtype = torch.long)
        for ns_idx, ns in enumerate(nodes):
            ng_eid = ng_sid + ns.num_node_blocks

            n_chgs[ng_sid:ng_eid] = ns.num_chs

            layer_num_edges += ns.num_nodes * ns.num_chs

            ns._output_ind_range = (global_nid_start, global_nid_start + ns.num_nodes)
            global_nid_start += ns.num_nodes

            ng_sid = ng_eid
    else:
        layer_num_nblock = sum(map(lambda ns: ns.num_nodes, nodes))
        layer_num_edges = 0

        ng_sid = 0
        n_chgs = torch.zeros([layer_num_nblock], dtype = torch.long)
        for ns_idx, ns in enumerate(nodes):
            ng_eid = ng_sid + ns.num_nodes

            n_chgs[ng_sid:ng_eid] = ns.num_chs

            layer_num_edges += ns.num_nodes * ns.num_chs

            ns._output_ind_range = (global_nid_start, global_nid_start + ns.num_nodes)
            global_nid_start += ns.num_nodes

            ng_sid = ng_eid

    return layer_num_nblock, layer_num_edges, n_chgs


@torch.no_grad()
def prod_layer_forward_compilation(nodes, fw_partition_max_chs, n_partition_ids, n_id_in_partition, num_ngs_in_partition, 
                                   block_size, use_block_sparse_edges: bool, use_cuda: bool = False):
    
    if use_cuda and not torch.cuda.is_available():
        use_cuda = False

    if not use_block_sparse_edges:
        assert block_size == 1

    if use_cuda:
        device = torch.device(f"cuda:{torch.cuda.current_device()}")
    else:
        device = torch.device("cpu")

    nids = [torch.zeros([partition_size], dtype = torch.long, device = device) for partition_size in num_ngs_in_partition] # Node block start id
    cids = [torch.zeros([partition_size, max_chs] , dtype = torch.long, device = device) for partition_size, max_chs in zip(num_ngs_in_partition, fw_partition_max_chs)] # Child block start id

    for ns_id, ns in enumerate(nodes):

        # `partition_id`:   which partition the current node belongs to
        # `local_sid`:      the start index of the node within the current partition
        # `partition_nchs`: maximum number of child nodes in the current partition
        partition_id = n_partition_ids[ns_id]
        local_sid = n_id_in_partition[ns_id]
        if use_block_sparse_edges:
            local_eid = local_sid + ns.num_node_blocks
        else:
            local_eid = local_sid + ns.num_nodes
        partition_nchs = fw_partition_max_chs[partition_id]

        if use_block_sparse_edges:
            n_sid = ns._output_ind_range[0]
            nids[partition_id][local_sid:local_eid] = torch.arange(0, ns.num_nodes, block_size, device = device) + n_sid
            for cs_id, cs in enumerate(ns.chs):
                cids[partition_id][local_sid:local_eid,cs_id] = ns.edge_ids[:,cs_id].to(device) * block_size + cs._output_ind_range[0]
        else:
            n_sid = ns._output_ind_range[0]
            nids[partition_id][local_sid:local_eid] = torch.arange(0, ns.num_nodes, device = device) + n_sid
            if ns.is_sparse():
                for cs_id, cs in enumerate(ns.chs):
                    cids[partition_id][local_sid:local_eid,cs_id] = ns.edge_ids[:,cs_id].to(device) + cs._output_ind_range[0]
            else:
                assert ns.is_block_sparse()
                edge_ids = ns.edge_ids.clone()
                edge_ids = (edge_ids[:,None,:].repeat(1, ns.block_size, 1) * ns.block_size + torch.arange(0, ns.block_size)[None,:,None]).flatten(0, 1)
                for cs_id, cs in enumerate(ns.chs):
                    cids[partition_id][local_sid:local_eid,cs_id] = edge_ids[:,cs_id].to(device) + cs._output_ind_range[0]

    if use_cuda:
        nids = [tensor.cpu() for tensor in nids]
        cids = [tensor.cpu() for tensor in cids]

    return nids, cids


@torch.no_grad()
def flatten_c_ids(nids: Sequence[torch.Tensor], cids: Sequence[torch.Tensor]):

    num_cid_slots = sum(map(lambda x: x.size(0) * x.size(1), cids))
    flat_cids = torch.zeros([num_cid_slots], dtype = torch.long)
    flat_cid2nid = torch.zeros([num_cid_slots], dtype = torch.long)

    n_sid = 0
    c_sid = 0
    for curr_nids, curr_cids in zip(nids, cids):
        n_eid = n_sid + curr_nids.size(0)
        c_eid = c_sid + curr_cids.size(0) * curr_cids.size(1)

        flat_cids[c_sid:c_eid] = curr_cids.reshape(-1)
        flat_cid2nid[c_sid:c_eid] = curr_nids.unsqueeze(1).repeat(1, curr_cids.size(1)).reshape(-1)

        n_sid = n_eid
        c_sid = c_eid

    return flat_cids, flat_cid2nid


@torch.no_grad()
def get_prod_layer_parstats(flat_cids: torch.Tensor, global_nid_start: int):

    u_cids, par_counts = torch.unique(flat_cids, sorted = True, return_counts = True)

    c_sids = torch.arange(0, u_cids.size(0))[u_cids < global_nid_start]

    if c_sids.numel() > 0:
        # Strip away dummy nodes
        c_sid = c_sids.max() + 1
        u_cids = u_cids[c_sid:]
        par_counts = par_counts[c_sid:]

    return u_cids, par_counts


@njit
def _assign_c_idx_kernel(flat_cids, flat_par_offsets, cum_c_nodes):
    for i in range(flat_cids.shape[0]):
        cid = flat_cids[i]
        idx = cum_c_nodes[cid]
        flat_par_offsets[i] = idx
        cum_c_nodes[cid] = idx + 1


@njit
def _assign_cid2_block_local_id(flat_u_cids, n_block_ids, n_id_in_block, cid2block_id, cid2local_id):
    for i in range(flat_u_cids.shape[0]):
        cid = flat_u_cids[i]
        cid2block_id[cid] = n_block_ids[i]
        cid2local_id[cid] = n_id_in_block[i]


@triton.jit
def _assign_target_ucids_kernel(target_u_cids_ptr, flat_u_cids_ptr, n_partition_ids_ptr, n_id_in_partition_ptr, 
                                u_cids_partition_start_ptr, constexprs_ptr, BLOCK_SIZE: tl.constexpr):

    pid = tl.program_id(axis = 0)
    block_start = pid * BLOCK_SIZE

    # Retrieve all constexprs
    num_nodes = tl.load(constexprs_ptr)

    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < num_nodes

    # Get `cid`
    cid = tl.load(flat_u_cids_ptr + offsets, mask = mask, other = 0)

    # Get `partition_id` and `local_id`
    partition_id = tl.load(n_partition_ids_ptr + offsets, mask = mask, other = 0)
    local_id = tl.load(n_id_in_partition_ptr + offsets, mask = mask, other = 0)

    # Get the corresponding start id in the target tensors
    u_cids_start = tl.load(u_cids_partition_start_ptr + partition_id, mask = mask, other = 0)

    # Assign to `target_u_cids`
    tl.store(target_u_cids_ptr + u_cids_start + local_id, cid, mask = mask)


@triton.jit
def _assign_prod_target_parids_kernel(target_parids_ptr, flat_cid2nid_ptr, flat_cids_ptr, 
                                      cid2partition_id_ptr, cid2local_id_ptr, parids_partition_start_ptr,
                                      flat_par_offsets_ptr, bk_partition_max_pars_ptr, constexprs_ptr, 
                                      BLOCK_SIZE: tl.constexpr):

    pid = tl.program_id(axis = 0)
    block_start = pid * BLOCK_SIZE

    # Retrieve all constexprs
    num_edges = tl.load(constexprs_ptr)

    offsets = block_start + tl.arange(0, BLOCK_SIZE)
    mask = offsets < num_edges

    # Get `nid` and `cid` of the edges
    nid = tl.load(flat_cid2nid_ptr + offsets, mask = mask, other = 0)
    cid = tl.load(flat_cids_ptr + offsets, mask = mask, other = 0)

    # Mask out edges that point to the dummy node
    mask = mask & (cid != 0)

    # Get `partition_id` and `local_id` using `cid`
    partition_id = tl.load(cid2partition_id_ptr + cid, mask = mask, other = 0)
    local_id = tl.load(cid2local_id_ptr + cid, mask = mask, other = 0)

    # Get the corresponding start id in the target tensors
    parids_start = tl.load(parids_partition_start_ptr + partition_id, mask = mask, other = 0)

    # Get `par_offset` of the edges
    par_offset = tl.load(flat_par_offsets_ptr + offsets, mask = mask, other = 0)

    # Assign to `target_parids`
    partition_max_n_pars = tl.load(bk_partition_max_pars_ptr + partition_id, mask = mask, other = 0)
    parid_offsets = parids_start + local_id * partition_max_n_pars + par_offset
    tl.store(target_parids_ptr + parid_offsets, nid, mask = mask)


@torch.no_grad()
def prod_layer_backward_compilation(flat_u_cids, flat_cids, flat_cid2nid, bk_partition_max_pars, n_partition_ids, 
                                    n_id_in_partition, num_ns_in_partition, use_cuda: bool = False):
    
    if use_cuda and not torch.cuda.is_available():
        use_cuda = False

    if use_cuda:

        # We construct a flattened version of `u_cids` where the vectors of every partition is concatenated
        # into a single vector. `u_cids_partition_start` is used to indicate the start index of every partition's
        # `u_cids`. That is, `target_u_cids[u_cids_partition_start[gid]:u_cids_partition_start[gid+1]] == u_cids[gid]`
        u_cids_partition_start = torch.zeros_like(num_ns_in_partition)
        u_cids_partition_start[1:] = torch.cumsum(num_ns_in_partition[:-1], dim = 0)
        target_u_cids = torch.zeros([num_ns_in_partition.sum()], dtype = torch.long)

        # Similar to `target_u_cids`, we construct a flattened version of `parids` and use `parids_partition_start`
        # for indexing
        parids_partition_start = torch.zeros_like(num_ns_in_partition)
        parids_partition_start[1:] = torch.cumsum((num_ns_in_partition * bk_partition_max_pars)[:-1], dim = 0)
        target_parids = torch.zeros([(num_ns_in_partition * bk_partition_max_pars).sum()], dtype = torch.long)

        # Precompute the parent offset ids for every edge. That is, the `?` mark in `parids[partition_id][local_id,?]`
        flat_par_offsets = np.zeros([flat_cids.size(0)], dtype = np.int64)
        num_c_nblocks = flat_u_cids.max().item() + 1
        cum_c_nblocks = np.zeros([num_c_nblocks], dtype = np.int64)

        _assign_c_idx_kernel(flat_cids.numpy(), flat_par_offsets, cum_c_nblocks)
        flat_par_offsets = torch.from_numpy(flat_par_offsets).cuda()

        # Direct mapping from `cid` to `block_id` and `local_id`
        cid2partition_id = np.zeros([num_c_nblocks], dtype = np.int64)
        cid2local_id = np.zeros([num_c_nblocks], dtype = np.int64)

        _assign_cid2_block_local_id(flat_u_cids.numpy(), n_partition_ids.numpy(), n_id_in_partition.numpy(), cid2partition_id, cid2local_id)
        cid2partition_id = torch.from_numpy(cid2partition_id).cuda()
        cid2local_id = torch.from_numpy(cid2local_id).cuda()

        # The following kernel assigns the indices to `target_u_cids` and `target_parids`. This is equivalent
        # to the easier-to-read CPU version enabled by setting `use_cuda = False`
        num_nblocks = flat_u_cids.size(0)
        num_edges = flat_cids.size(0)
        flat_u_cids = flat_u_cids.cuda()
        n_partition_ids = n_partition_ids.cuda()
        n_id_in_partition = n_id_in_partition.cuda()
        target_u_cids = target_u_cids.cuda()
        target_parids = target_parids.cuda()
        flat_cid2nid = flat_cid2nid.cuda()
        flat_cids = flat_cids.cuda()
        u_cids_partition_start = u_cids_partition_start.cuda()
        parids_partition_start = parids_partition_start.cuda()
        bk_partition_max_pars = bk_partition_max_pars.cuda()

        # We store these constants in a tensor and retrieve them in the kernel
        constexprs1 = torch.tensor([num_nblocks]).long().cuda()
        constexprs2 = torch.tensor([num_edges]).long().cuda()

        grid1 = lambda meta: (triton.cdiv(num_nblocks, meta["BLOCK_SIZE"]),)

        _assign_target_ucids_kernel[grid1](
            target_u_cids, flat_u_cids, n_partition_ids, n_id_in_partition, 
            u_cids_partition_start, constexprs1, BLOCK_SIZE = 2048
        )

        grid2 = lambda meta: (triton.cdiv(num_edges, meta["BLOCK_SIZE"]),)

        _assign_prod_target_parids_kernel[grid2](
            target_parids, flat_cid2nid, flat_cids, 
            cid2partition_id, cid2local_id, parids_partition_start,
            flat_par_offsets, bk_partition_max_pars, constexprs2, BLOCK_SIZE = 2048
        )

        target_u_cids = target_u_cids.cpu()
        u_cids = []
        for partition_id in range(num_ns_in_partition.size(0)):
            sid = u_cids_partition_start[partition_id]
            eid = sid + num_ns_in_partition[partition_id]
            u_cids.append(target_u_cids[sid:eid].contiguous())

        target_parids = target_parids.cpu()
        parids = []
        for partition_id in range(num_ns_in_partition.size(0)):
            sid = parids_partition_start[partition_id]
            psize = num_ns_in_partition[partition_id]
            pnpar = bk_partition_max_pars[partition_id]
            eid = sid + psize * pnpar
            parids.append(target_parids[sid:eid].reshape(psize, pnpar).contiguous())

    else:

        u_cids = [torch.zeros([partition_size], dtype = torch.long) for partition_size in num_ns_in_partition] # Node block id
        parids = [torch.zeros([partition_size, max_n_pars], dtype = torch.long) for partition_size, max_n_pars in zip(num_ns_in_partition, bk_partition_max_pars)] # Parent block id

        for idx in range(flat_u_cids.size(0)):
            cid = flat_u_cids[idx]

            # `partition_id`:   which partition the current node block belongs to
            # `local_id`:       the index of the node block within the current partition
            partition_id = n_partition_ids[idx]
            local_id = n_id_in_partition[idx]

            criterion = (flat_cids == cid)
            npar = criterion.sum()

            u_cids[partition_id][local_id] = cid
            parids[partition_id][local_id,:npar] = flat_cid2nid[criterion]

    return u_cids, parids



from __future__ import annotations

import torch
import torch.nn as nn

from typing import Sequence, Callable, Optional

from .layer import Layer
from .input_layer import InputLayer
from .prod_layer import ProdLayer
from .sum_layer import SumLayer


class LayerGroup(nn.Module):
    def __init__(self, layers: Sequence[Layer]):
        super(LayerGroup, self).__init__()

        assert len(layers) >= 1, "A `LayerGroup` must contains at least 1 layer."

        for i in range(1, len(layers)):
            assert type(layers[i]) == type(layers[0])

        if isinstance(layers[0], InputLayer):
            self.layer_type = "input"
        elif isinstance(layers[0], ProdLayer):
            self.layer_type = "prod"
        else:
            assert isinstance(layers[0], SumLayer)
            self.layer_type = "sum"

        self.num_layers = len(layers)

        self.layers = []
        for i, layer in enumerate(layers):
            self.add_module(f"layer_{i}", layer)
            self.layers.append(layer)

    def to(self, device):
        super(LayerGroup, self).to(device)

        for layer in self.layers:
            layer.to(device)

    def forward(self, *args, **kwargs):

        for layer in self.layers:
            layer.forward(*args, **kwargs)

    def backward(self, *args, **kwargs):

        for layer in self.layers:
            layer.backward(*args, **kwargs)

    def callback(self, callback_fn: Optional[Callable], *args, **kwargs):

        if callback_fn is not None:
            for layer in self.layers:
                callback_fn(layer, *args, **kwargs)

    def enable_partial_evaluation(self, *args, **kwargs):

        for layer in self.layers:
            layer.enable_partial_evaluation(*args, **kwargs)

    def disable_partial_evaluation(self, *args, **kwargs):

        for layer in self.layers:
            layer.disable_partial_evaluation(*args, **kwargs)

    def is_input(self):
        return self.layer_type == "input"
        
    def is_prod(self):
        return self.layer_type == "prod"

    def is_sum(self):
        return self.layer_type == "sum"

    def __len__(self):
        self.num_layers

    def __getitem__(self, idx):
        return self.layers[idx]

    def __iter__(self):
        self.iter_idx = 0
        return self

    def __next__(self):
        if self.iter_idx < self.num_layers:
            layer = self.layers[self.iter_idx]
            self.iter_idx += 1
            return layer
        else:
            raise StopIteration

    def _prepare_scope2nids(self, *args, **kwargs):
        
        if self.is_prod():
            prod_scope_eleids = list()
            for layer in self.layers:
                prod_scope_eleids.extend(layer._prepare_scope2nids(*args, **kwargs))

            return prod_scope_eleids
        
        else:
            for layer in self.layers:
                layer._prepare_scope2nids(*args, **kwargs)



from __future__ import annotations

import torch
import numpy as np
import random
import networkx as nx
from typing import Type, Optional

from pyjuice.nodes.distributions import *
from pyjuice.nodes import multiply, summate, inputs, set_block_size, CircuitNodes
from pyjuice.utils.util import max_cdf_power_of_2
from pyjuice.utils import BitSet


def RAT_SPN(num_vars: int, num_latents: int, depth: int, num_repetitions: int, num_pieces: int = 2,
            input_dist: Optional[Distribution] = None,
            input_node_type: Type[Distribution] = Categorical, 
            input_node_params: dict = {"num_cats": 256},
            block_size: Optional[int] = None):
    """
    Generate Random and Tensorized SPNs (https://proceedings.mlr.press/v115/peharz20a/peharz20a.pdf)

    :param num_vars: number of variables
    :type num_vars: int

    :param num_latents: size of the latent space
    :type num_latents: int

    :param depth: splitting depth of variable scopes
    :type depth: int

    :param num_repetitions: number of random splits
    :type num_repetitions: int

    :param num_pieces: the number of sub-scopes splitted from any scope
    :type num_pieces: int

    :param input_dist: input distribution
    :type input_dist: Distribution

    :param block_size: block size
    :type block_size: int
    """

    # Specify block size
    if block_size is None:
        block_size = min(256, max_cdf_power_of_2(num_latents))

    if input_dist is not None:
        input_node_type, input_node_params = input_dist._get_constructor()

    assert num_latents % block_size == 0, f"`num_latents` ({num_latents}) not divisible by `block_size` ({block_size})."
    num_node_blocks = num_latents // block_size

    with set_block_size(block_size):

        # Input nodes
        input_ns = []
        for v in range(num_vars):
            ns = inputs(v, num_node_blocks = num_node_blocks, dist = input_node_type(**input_node_params))
            input_ns.append(ns)

        # Top-down partition
        def partition_ns(scope, curr_depth = 0):
            if curr_depth >= depth or len(scope) < num_pieces:
                chs = [input_ns[v] for v in scope]
                np = multiply(*chs)
                ns = summate(np, num_node_blocks = num_node_blocks)

                return ns

            scope_list = scope.to_list()
            random.shuffle(scope_list)

            quotient, remainder = len(scope) // num_pieces, len(scope) % num_pieces
            list_lengths = [quotient + (i < remainder) for i in range(num_pieces)]

            ch_scopes = []
            sid = 0
            for i in range(num_pieces):
                eid = sid + list_lengths[i]
                ch_scopes.append(BitSet.from_array(scope_list[sid:eid]))
                sid = eid

            chs = [partition_ns(scope, curr_depth + 1) for scope in ch_scopes]
            np = multiply(*chs)
            ns = summate(np, num_node_blocks = num_node_blocks)

            return ns

        root_nps = []
        for rep_id in range(num_repetitions):
            root_ns = partition_ns(BitSet.from_array([v for v in range(num_vars)]))
            root_nps.append(root_ns.chs[0])

        root_ns = summate(*root_nps, num_node_blocks = 1, block_size = 1)

    return root_ns
    


from __future__ import annotations

import numpy as np
import torch
from typing import Sequence, Union, Optional, Callable
from copy import deepcopy
from collections import deque

from pyjuice.utils import BitSet
from pyjuice.graph import RegionGraph, PartitionNode, InnerRegionNode, InputRegionNode


def node_iterator(root_ns: CircuitNodes, reverse: bool = False):
    def dfs(ns: CircuitNodes, fn: Callable, visited: set = set()):
        if ns in visited:
            return

        visited.add(ns)

        # Recursively traverse children
        if ns.is_sum() or ns.is_prod():
            for cs in ns.chs:
                dfs(cs, fn = fn, visited = visited)

        fn(ns)

    if not reverse:
        visited = set()
        node_list = list()

        def record_fn(ns):
            node_list.append(ns)

        dfs(root_ns, record_fn)

        for ns in node_list:
            yield ns
    
    else:
        parcount = dict()
        node_list = list()

        def inc_parcount(ns):
            for cs in ns.chs:
                if cs not in parcount:
                    parcount[cs] = 0
                parcount[cs] += 1

        dfs(root_ns, inc_parcount)

        queue = deque()
        queue.append(root_ns)
        while len(queue) > 0:
            ns = queue.popleft()
            node_list.append(ns)
            for cs in ns.chs:
                parcount[cs] -= 1
                if parcount[cs] == 0:
                    queue.append(cs)

        assert len(parcount) + 1 == len(node_list)

        for ns in node_list:
            yield ns


class CircuitNodes():

    # A list of function that will be called at the end of `__init__`
    # This should only be changed by context managers, so please do not 
    # add anything here.
    INIT_CALLBACKS = []

    # Default `block_size`. Used by the context managers.
    DEFAULT_BLOCK_SIZE = 1

    def __init__(self, num_node_blocks: int, region_node: RegionGraph, block_size: int = 0, source_node: Optional[CircuitNodes] = None, **kwargs):

        if block_size == 0:
            block_size = self.DEFAULT_BLOCK_SIZE

        assert num_node_blocks > 0
        assert block_size > 0 and (block_size & (block_size - 1)) == 0, f"`block_size` must be a power of 2, but got `block_size={block_size}`."
        
        self.num_node_blocks = num_node_blocks
        self.block_size = block_size
        self.region_node = region_node

        self.chs = []
        
        self._output_ind_range = None
        self._param_ids = None
        self._params = None

        # Source nodes it points to (for parameter tying)
        if source_node is not None:
            while source_node._source_node is not None:
                source_node = source_node._source_node
        self._source_node = source_node

        self._tied_param_block_ids = None

        self._reverse_iter = False

    def _run_init_callbacks(self, **kwargs):
        for func in self.INIT_CALLBACKS:
            func(self, **kwargs)

    @property
    def scope(self):
        return self.region_node.scope

    def is_sum(self):
        return isinstance(self.region_node, InnerRegionNode)

    def is_prod(self):
        return isinstance(self.region_node, PartitionNode)

    def is_input(self):
        return isinstance(self.region_node, InputRegionNode)

    @property
    def num_chs(self):
        return len(self.chs)

    @property
    def num_nodes(self):
        """
        Number of PC nodes within the current node.
        """
        return self.num_node_blocks * self.block_size

    @property
    def num_edges(self):
        raise NotImplementedError()

    def duplicate(self, *args, **kwargs):
        raise ValueError(f"{type(self)} does not support `duplicate`.")

    def init_parameters(self, perturbation: float = 2.0, recursive: bool = True, visited: set = set(), 
                        is_root = True, **kwargs):
        if recursive:
            if self in visited:
                return None
            
            visited.add(self)

            for cs in self.chs:
                cs.init_parameters(
                    perturbation = perturbation, 
                    recursive = recursive, 
                    visited = visited, 
                    is_root = False,
                    **kwargs
                )
        else:
            visited.add(self)

        # Process all tied nodes
        for ns in visited:
            if ns._source_node is not None:
                # Do not store parameters explicitly for tied nodes
                # We can always retrieve them from the source nodes when required
                ns._params = None

    def num_parameters(self, count_input_params: bool = False, recursive: bool = True):
        if recursive:
            param_count = 0
            for ns in self:
                if ns.is_sum() and not ns.is_tied():
                    param_count += ns.num_edges
                elif ns.is_input() and count_input_params:
                    param_count += ns.num_nodes * ns.dist.num_parameters()

            return param_count
        else:
            if ns.is_sum() and not ns.is_tied():
                return ns.num_edges
            elif ns.is_input() and count_input_params:
                return ns.num_nodes * ns.dist.num_parameters()
            else:
                return 0

    def is_tied(self):
        return self._source_node is not None

    def get_source_ns(self):
        return self._source_node if self.is_tied() else self

    def set_source_ns(self, source_ns: CircuitNodes):
        assert type(source_ns) == type(self), f"Node type of the source ns ({type(source_ns)}) does not match that of self ({type(self)})."
        assert len(source_ns.chs) == len(self.chs), "Number of children does not match."
        assert not hasattr(self, "_params") or self._params is None, "The current node should not have parameters to avoid confusion."
        assert source_ns.num_node_blocks == self.num_node_blocks, "`num_node_blocks` does not match."
        assert source_ns.block_size == self.block_size,  "`block_size` does not match."

        self._source_node = source_ns

    def has_params(self):
        if self.is_input():
            return self._param_initialized

        if not self.is_tied():
            return hasattr(self, "_params") and self._params is not None
        else:
            source_ns = self.get_source_ns()
            return hasattr(source_ns, "_params") and source_ns._params is not None

    def contains(self, ns: CircuitNodes):
        for n in self:
            if n == ns:
                return True
        
        return False

    def _clear_tensor_circuit_hooks(self, recursive: bool = True):

        def clear_hooks(ns):
            if hasattr(ns, "_param_range"):
                ns._param_range = None
            if hasattr(ns, "_param_ids"):
                ns._param_ids = None
            if hasattr(ns, "_inverse_param_ids"):
                ns._inverse_param_ids = None
            if hasattr(ns, "_param_flow_range"):
                ns._param_flow_range = None
            if hasattr(ns, "_output_ind_range"):
                ns._output_ind_range = None

        if recursive:
            for ns in self:
                clear_hooks(ns)
        else:
            clear_hooks(self)

    def __len__(self):
        count = 0

        def dfs(ns: CircuitNodes, visited: set = set()):
            nonlocal count

            if ns in visited:
                return

            visited.add(ns)

            # Recursively traverse children
            if ns.is_sum() or ns.is_prod():
                for cs in ns.chs:
                    dfs(cs, visited = visited)

            count += 1

        dfs(self)
        return count

    def __iter__(self):
        return node_iterator(self, self._reverse_iter)

    def __call__(self, reverse: bool = False):
        self._reverse_iter = reverse

        return self

    def provided(self, var_name):
        return hasattr(self, var_name) and getattr(self, var_name) is not None



from __future__ import annotations

import torch
import triton
import triton.language as tl
from typing import Union, Callable
from functools import partial

from pyjuice.nodes import CircuitNodes
from pyjuice.model import TensorCircuit
from .base import query

from .conditional import _conditional_fw_input_fn


def marginal(pc: TensorCircuit, data: torch.Tensor, missing_mask: Optional[torch.Tensor] = None, 
             fw_input_fn: Optional[Union[str,Callable]] = None, **kwargs):
    """
    Compute the marginal probability given the assignment of a subset of variables, i.e., P(e).

    :param pc: the input PC
    :type pc: TensorCircuit

    :param data: data of size [B, num_vars] (hard evidence) or a custom shape paired with `fw_input_fn`
    :type data: torch.Tensor

    :param missing_mask: a boolean mask indicating marginalized variables; the size can be [num_vars] or [B, num_vars]
    :type missing_mask: torch.Tensor

    :param fw_input_fn: an optional custom function for the forward pass of input layers
    :type fw_input_fn: Optional[Union[str,Callable]]
    """
    
    lls = query(pc, inputs = data, run_backward = False, 
                fw_input_fn = _conditional_fw_input_fn if fw_input_fn is None else fw_input_fn,
                missing_mask = missing_mask, **kwargs)

    return lls